#    VenviPy - A Virtual Environment Manager for Python.
#    Copyright (C) 2021 - Youssef Serestou - sinusphi.sq@gmail.com
#
#    This program is free software: you can redistribute it and/or modify
#    it under the terms of the GNU General Public License as published by
#    the Free Software Foundation, either version 3 of the License or any
#    later version.
#
#    This program is distributed in the hope that it will be useful,
#    but WITHOUT ANY WARRANTY; without even the implied warranty of
#    MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
#    GNU General Public License for more details.
#
#    A copy of the GNU General Public License version 3 named LICENSE is
#    in the root directory of VenviPy.
#    If not, see <https://www.gnu.org/licenses/licenses.en.html#GPL>.

# -*- coding: utf-8 -*-
"""
This module contains the package manager.
"""
import os
import logging
import webbrowser

from PyQt6.QtGui import (
    QIcon,
    QCursor,
    QPixmap,
    QStandardItem,
    QStandardItemModel,
    QAction
)
from PyQt6.QtCore import Qt, pyqtSignal, pyqtSlot
from PyQt6.QtWidgets import (
    QFileDialog,
    QDialog,
    QHBoxLayout,
    QLabel,
    QLineEdit,
    QApplication,
    QPushButton,
    QGridLayout,
    QMessageBox,
    QAbstractItemView,
    QStyle,
    QTableView,
    QMenu,
    QFrame
)

import venvipy_rc  # pylint: disable=unused-import
import get_data
import creator
from manage_pip import PipManager
from dialogs import ConsoleDialog
from styles.theme import PACKAGE_DIALOG_QSS
from styles.custom import (
    package_manager_title_text,
    package_manager_venv_name_text,
)

logger = logging.getLogger(__name__)



class PackagesTable(QTableView):
    """Contains the Packages installed in the selected venv.
    """
    context_triggered = pyqtSignal()
    refresh_packages = pyqtSignal()

    def __init__(self, *args, **kwargs):
        super().__init__(*args, **kwargs)

        self.delete_icon = QIcon(
            self.style().standardIcon(QStyle.StandardPixmap.SP_TrashIcon)
        )
        self.info_icon = QIcon(
            self.style().standardIcon(QStyle.StandardPixmap.SP_FileDialogInfoView)
        )

        self.console = ConsoleDialog()


    def get_selected_item(self):
        """
        Return `str` from `name` column of the selected row.
        """
        listed_items = self.selectionModel().selectedRows()
        for index in listed_items:
            selected_item = index.data()
            return selected_item


    def contextMenuEvent(self, event):

        idx = self.indexAt(event.pos())

        if not idx.isValid():
            idx = self.currentIndex()

            if not idx.isValid():
                return

        self.selectRow(idx.row())

        menu = QMenu(self)

        install_action = QAction(QIcon.fromTheme("software-install"), "&Install module", self)
        install_action.triggered.connect(self.context_triggered.emit)
        menu.addAction(install_action)

        open_pypi_action = QAction(self.info_icon, "&Open on PyPI", self)
        open_pypi_action.triggered.connect(lambda: self.open_on_pypi(event))
        menu.addAction(open_pypi_action)

        pos = event.globalPos()
        if pos.x() == 0 and pos.y() == 0:
            pos = self.viewport().mapToGlobal(self.visualRect(idx).center())

        menu.exec(pos)


    def open_on_pypi(self, event):
        """
        Open pypi.org and show the project page
        of the selected package.
        """
        url = "https://pypi.org/project"
        package = self.get_selected_item()
        webbrowser.open("/".join([url, package]))


    def update_package(self, event):
        """Run pip --update 'package'.
        """
        self.setEnabled(False)

        # get the venv path
        with open(get_data.ACTIVE_VENV, "r", encoding="utf-8") as f:
            venv_path = f.read()

        split_venv_path = os.path.split(venv_path)
        self.venv_location = split_venv_path[0]
        self.venv_name = os.path.basename(venv_path)
        self.pkg = self.get_selected_item()

        msg_box_question = QMessageBox.question(
            self,
            "Confirm", f"Are you sure you want to update '{self.pkg}'?      ",
            QMessageBox.StandardButton.Cancel | QMessageBox.StandardButton.Yes
        )

        if msg_box_question == QMessageBox.StandardButton.Yes:
            self.console.setWindowTitle(f"Updating {self.pkg}")

            self.manager = PipManager(
                self.venv_location,
                self.venv_name
            )
            # open the console when recieving signal from manager
            self.manager.started.connect(self.console.exec)
            self.manager.text_changed.connect(self.console.update_status)

            # start installing the selected package
            logger.debug(f"Updating '{self.pkg}'...")
            self.manager.run_pip(creator.cmds[0], [creator.opts[0], self.pkg])

            # clear the content when closing console
            if self.console.close:
                self.console.console_window.clear()

        self.setEnabled(True)



class PackageManager(QDialog):
    """
    The package manager dialog.
    """
    refresh_packages = pyqtSignal()

    def __init__(self, *args, **kwargs):
        super().__init__(*args, **kwargs)

        self.setWindowTitle("Package Manager")
        self.resize(975, 650)
        self.center()
        self.setWindowIcon(QIcon(":/img/profile.png"))

        reload_icon = QIcon(
            self.style().standardIcon(QStyle.StandardPixmap.SP_BrowserReload)
        )

        # get the venv path from file
        try:
            with open(get_data.ACTIVE_VENV, "r", encoding="utf-8") as f:
                venv_path = f.read()
        except FileNotFoundError:
            venv_path = ""

        split_venv_path = os.path.split(venv_path)
        self.venv_location = split_venv_path[0]
        self.venv_name = os.path.basename(venv_path)

        self.setStyleSheet(PACKAGE_DIALOG_QSS)


        #]===================================================================[#
        #] PAGE CONTENT [#===================================================[#
        #]===================================================================[#

        horizontal_layout = QHBoxLayout(self)
        horizontal_layout.setContentsMargins(25, 18, 25, 22)

        grid_layout = QGridLayout()

        title_label = QLabel(package_manager_title_text)

        logo = QLabel()
        pixmap = QPixmap(":/img/pypi.png")
        logo_scaled = pixmap.scaled(
            92, 92, Qt.AspectRatioMode.KeepAspectRatio
        )
        logo.setPixmap(logo_scaled)

        venv_name_colored = package_manager_venv_name_text(self.venv_name)

        subtitle_label_1 = QLabel(
            f"Listet below are the Python packages installed in {venv_name_colored} "
        )
        subtitle_label_1.setContentsMargins(22, 0, 0, 0)

        subtitle_label_2 = QLabel(
            "Select one and right-click for available options. "
        )
        subtitle_label_2.setContentsMargins(22, 0, 0, 0)

        line_1 = QFrame(self)
        line_1.setFixedHeight(15)
        line_1.setFrameShape(QFrame.Shape.HLine)
        line_1.setFrameShadow(QFrame.Shadow.Sunken)

        pkg_name_label = QLabel("Package &name:")
        self.pkg_name_line = QLineEdit()
        pkg_name_label.setBuddy(self.pkg_name_line)

        self.reload_button = QPushButton(
            reload_icon,
            "&Reload",
            clicked=self.pop_packages_table
        )

        exit_button = QPushButton(
            "&Exit",
            clicked=self.on_close
        )

        # packages table
        self.packages_table = PackagesTable(
            selectionBehavior=QAbstractItemView.SelectionBehavior.SelectRows,
            editTriggers=QAbstractItemView.EditTrigger.NoEditTriggers,
            alternatingRowColors=True,
            sortingEnabled=True,
            #refresh_packages=self.pop_packages_table
            #context_triggered=self.install_package
            ### TODO: add signal
        )

        # hide vertical header
        self.packages_table.verticalHeader().hide()

        # adjust horizontal headers
        h_header = self.packages_table.horizontalHeader()
        h_header.setDefaultAlignment(Qt.AlignmentFlag.AlignLeft)
        h_header.setStretchLastSection(True)

        # set table view model
        self.packages_table_model = QStandardItemModel(0, 3, self)
        self.packages_table.setModel(self.packages_table_model)

        line_2 = QFrame(self)
        line_2.setFixedHeight(8)
        line_2.setFrameShape(QFrame.Shape.HLine)
        line_2.setFrameShadow(QFrame.Shadow.Sunken)

        grid_layout.addWidget(title_label, 0, 0, 1, 2)
        grid_layout.addWidget(logo, 0, 2, 1, 1)
        grid_layout.addWidget(subtitle_label_1, 1, 0, 1, 2)
        grid_layout.addWidget(subtitle_label_2, 2, 0, 1, 2)
        grid_layout.addWidget(line_1, 3, 0, 1, 3)
        grid_layout.addWidget(pkg_name_label, 4, 0, 1, 1)
        grid_layout.addWidget(self.pkg_name_line, 4, 1, 1, 1)
        grid_layout.addWidget(self.reload_button, 4, 2, 1, 1)
        grid_layout.addWidget(self.packages_table, 5, 0, 1, 3)
        grid_layout.addWidget(line_2, 6, 0, 1, 3)
        grid_layout.addWidget(exit_button, 7, 2, 1, 1)

        horizontal_layout.addLayout(grid_layout)


    def center(self):
        """Center window."""
        qr = self.frameGeometry()
        screen = self.screen() or QApplication.primaryScreen()
        if screen:
            cp = screen.availableGeometry().center()
            qr.moveCenter(cp)
            self.move(qr.topLeft())


    def on_close(self):
        """Close window if no changes made, else ask to save a requirements.
        """
        self.close()


    def launch(self):
        """Launches the manager.
        """
        # count whether changes has been made to venv
        #self.venv_modified = 0

        # clear input
        self.packages_table_model.clear()
        self.pkg_name_line.clear()
        self.pkg_name_line.setFocus()

        # set text in column headers
        self.packages_table_model.setHorizontalHeaderLabels([
            "Name",
            "Version",
            "Author",
            "Description"
        ])

        # adjust column width
        self.packages_table.setColumnWidth(0, 200)  # name
        self.packages_table.setColumnWidth(1, 80)  # version
        self.packages_table.setColumnWidth(2, 150)  # release

        # connect to pop_packages_table method and call it
        self.refresh_packages.connect(self.pop_packages_table)
        self.refresh_packages.emit()

        # launch the dialog via exec_() method
        self.exec()


    def pop_packages_table(self):
        """Fill (refresh) the packages table content.
        """
        self.packages_table_model.setRowCount(0)

        for info in get_data.get_installed_packages(
            self.venv_location,
            self.venv_name
        ):
            self.packages_table_model.insertRow(0)

            for i, text in enumerate((
                info.pkg_name,
                info.pkg_version,
                info.pkg_info_2,
                info.pkg_summary
            )):
                self.packages_table_model.setItem(0, i, QStandardItem(text))



#    VenviPy - A Virtual Environment Manager for Python.
#    Copyright (C) 2021 - Youssef Serestou - sinusphi.sq@gmail.com
#
#    This program is free software: you can redistribute it and/or modify
#    it under the terms of the GNU General Public License as published by
#    the Free Software Foundation, either version 3 of the License or any
#    later version.
#
#    This program is distributed in the hope that it will be useful,
#    but WITHOUT ANY WARRANTY; without even the implied warranty of
#    MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
#    GNU General Public License for more details.
#
#    A copy of the GNU General Public License version 3 named LICENSE is
#    in the root directory of VenviPy.
#    If not, see <https://www.gnu.org/licenses/licenses.en.html#GPL>.

# -*- coding: utf-8 -*-
"""
Platform helpers for OS-specific behavior.
"""
import sys

from .linux import LinuxPlatform
from .windows import WindowsPlatform


def get_platform():
    if sys.platform.startswith("win"):
        return WindowsPlatform()
    return LinuxPlatform()


__all__ = ["get_platform", "LinuxPlatform", "WindowsPlatform"]



#    VenviPy - A Virtual Environment Manager for Python.
#    Copyright (C) 2021 - Youssef Serestou - sinusphi.sq@gmail.com
#
#    This program is free software: you can redistribute it and/or modify
#    it under the terms of the GNU General Public License as published by
#    the Free Software Foundation, either version 3 of the License or any
#    later version.
#
#    This program is distributed in the hope that it will be useful,
#    but WITHOUT ANY WARRANTY; without even the implied warranty of
#    MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
#    GNU General Public License for more details.
#
#    A copy of the GNU General Public License version 3 named LICENSE is
#    in the root directory of VenviPy.
#    If not, see <https://www.gnu.org/licenses/licenses.en.html#GPL>.

# -*- coding: utf-8 -*-
"""
This module provides several necessary data.
"""
import re
import os
import sys
import csv
import time
import json
import shutil
import sqlite3
import logging
from pathlib import Path
from typing import List, Optional, Dict, Any
from subprocess import PIPE, STDOUT, run
from dataclasses import dataclass

from bs4 import BeautifulSoup
import requests

from platforms import get_platform

__version__ = "0.4.2"

CFG_DIR = Path.home() / ".venvipy"
DB_FILE = Path.home() / ".venvipy" / "py-installs"
ACTIVE_DIR = Path.home() / ".venvipy" / "selected-dir"
ACTIVE_VENV = Path.home() / ".venvipy" / "active-venv"
TABS_STATE = Path.home() / ".venvipy" / "tabs-state.json"
PYPI_SIMPLE_URL = "https://pypi.org/simple/"
PYPI_JSON_URL = "https://pypi.org/pypi/{name}/json"
PACKAGE_DB_PATH = Path.home() / ".venvipy" / "pypi_index.sqlite3"
DB_TABLE = "projects"
DB_COL = "name"

logger = logging.getLogger(__name__)


#]===========================================================================[#
#] FIND PYTHON 3 INSTALLATIONS [#============================================[#
#]===========================================================================[#

@dataclass
class PythonInfo:
    """
    Info about Python installs.
    """
    py_version: str
    py_path: str


def to_version(value):
    """Convert a value to a readable version string.
    """
    return f"Python {value}"


def to_path(bin_path, version):
    """Return the absolute path to a python binary.
    """
    platform = get_platform()
    base_path = Path(bin_path)
    if platform.is_windows():
        return str(base_path / platform.python_exe_name())
    return str(base_path / f"{platform.python_exe_name()}{version}")


def is_writable(target_dir):
    """Test whether a directory is writable.
    """
    if os.path.exists(target_dir):
        test_file = os.path.join(target_dir, "test_file")

        try:
            logger.debug("Testing whether filesystem is writable...")
            with open(test_file, "w+", encoding="utf-8") as f:
                f.write("test")

            os.remove(test_file)
            logger.debug("Filesystem is writable")
            return True

        except OSError as e:
            logger.debug(f"Filesystem is read-only\n{e}")
            return False

    else:
        logger.debug(f"No such file or directory: {target_dir}")
        return False

    return False


def ensure_confdir():
    """Create `~/.venvipy` config directory.
    """
    if not os.path.exists(CFG_DIR):
        os.mkdir(CFG_DIR)


def ensure_dbfile():
    """Create the database in `~/.venvipy/py-installs`.
    """
    if not os.path.exists(DB_FILE):
        get_python_installs()


def ensure_active_dir():
    """Create the file that holds the selected path to venvs.
    """
    ensure_confdir()
    if not os.path.exists(ACTIVE_DIR):
        with open(ACTIVE_DIR, "w+", encoding="utf-8") as f:
            f.write("")


def ensure_active_venv():
    """Create the file that holds the selected path to venvs.
    """
    ensure_confdir()
    if not os.path.exists(ACTIVE_VENV):
        with open(ACTIVE_VENV, "w+", encoding="utf-8") as f:
            f.write("")


def default_tabs_state() -> Dict[str, Any]:
    """Return the default structure for stored tabs.
    """
    return {
        "tabs": [],
        "active_index": 0,
        "always_save_tabs": False,
        "ask_before_saving_tabs": True
    }


def load_tabs_state() -> Dict[str, Any]:
    """Load persisted tab information if present.
    """
    ensure_confdir()

    if not os.path.exists(TABS_STATE):
        return default_tabs_state()

    try:
        with open(TABS_STATE, "r", encoding="utf-8") as f:
            data = json.load(f)
            if isinstance(data, dict):
                return data
    except (OSError, json.JSONDecodeError):
        logger.warning("Could not read tabs state file; starting fresh")

    return default_tabs_state()


def save_tabs_state(state: Dict[str, Any]) -> None:
    """Persist tab information to disk.
    """
    ensure_confdir()

    try:
        with open(TABS_STATE, "w", encoding="utf-8") as f:
            json.dump(state, f, indent=2)
    except OSError as e:
        logger.warning(f"Failed to save tabs state: {e}")


def get_python_version(py_path):
    """Return Python version.
    """
    res = run(
        [py_path, "-V"],
        stdout=PIPE,
        stderr=STDOUT,
        encoding="utf-8",
        errors="replace",
        text=True,
        check=False,
    )

    python_version = (res.stdout or "").strip()
    return python_version


def get_python_installs(relaunching=False):
    """
    Write the found Python versions to `py-installs`. Create
    a new database if `relaunching=True`.
    """
    LATEST = 15
    versions = [f"3.{v}" for v in range(LATEST, 2, -1)]
    py_info_list = []
    platform = get_platform()

    ensure_confdir()

    if not os.path.exists(DB_FILE) or relaunching:
        with open(DB_FILE, "w", newline="", encoding="utf-8") as cf:
            fields = ["PYTHON_VERSION", "PYTHON_PATH"]
            writer = csv.DictWriter(
                cf,
                delimiter=",",
                quoting=csv.QUOTE_ALL,
                fieldnames=fields
            )
            writer.writeheader()

            if platform.is_windows():
                python_paths = _get_windows_python_paths()
                for python_path in python_paths:
                    if _is_venv_interpreter(python_path):
                        continue
                    python_version = get_python_version(python_path)
                    py_info = PythonInfo(python_version, python_path)
                    py_info_list.append(py_info)
                    writer.writerow({
                        "PYTHON_VERSION": py_info.py_version,
                        "PYTHON_PATH": py_info.py_path
                    })
            else:
                for version in versions:
                    python_path = shutil.which(f"python{version}")
                    if python_path is not None:
                        if _is_venv_interpreter(python_path):
                            continue
                        python_version = get_python_version(python_path)
                        py_info = PythonInfo(python_version, python_path)
                        py_info_list.append(py_info)
                        writer.writerow({
                            "PYTHON_VERSION": py_info.py_version,
                            "PYTHON_PATH": py_info.py_path
                        })
            cf.close()

            # add the system's Python manually if running in a virtual env
            if "VIRTUAL_ENV" in os.environ:
                system_python = os.path.realpath(sys.executable)
                add_python(system_python)

        return py_info_list[::-1]
    return False


def _get_windows_python_paths():
    paths = []
    py_launcher = run(
        ["py", "-0p"],
        stdout=PIPE,
        stderr=STDOUT,
        encoding="utf-8",
        errors="replace",
        text=True,
        check=False,
    )
    if py_launcher.returncode == 0:
        for line in py_launcher.stdout.splitlines():
            line = line.strip()
            if not line:
                continue
            parts = line.split()
            candidate = parts[-1] if parts else ""
            if candidate and os.path.exists(candidate):
                paths.append(candidate)

    if not paths:
        where_python = run(
            ["where", "python"],
            stdout=PIPE,
            stderr=STDOUT,
            encoding="utf-8",
            errors="replace",
            text=True,
            check=False,
        )
        if where_python.returncode == 0:
            for line in where_python.stdout.splitlines():
                candidate = line.strip()
                if candidate and os.path.exists(candidate):
                    paths.append(candidate)

    unique_paths = []
    for path in paths:
        if path not in unique_paths:
            unique_paths.append(path)
    return unique_paths


def add_python(py_path):
    """
    Write (append) Python version and path to `py-installs`.
    """
    ensure_dbfile()

    with open(DB_FILE, "a", newline="", encoding="utf-8") as cf:
        fields = ["PYTHON_VERSION", "PYTHON_PATH"]
        writer = csv.DictWriter(
            cf,
            delimiter=",",
            quoting=csv.QUOTE_ALL,
            fieldnames=fields
        )
        writer.writerow({
            "PYTHON_VERSION": get_python_version(py_path),
            "PYTHON_PATH": py_path
        })
        cf.close()

    # remove the interpreter if running in a virtual env
    if "VIRTUAL_ENV" in os.environ:
        remove_env()

def _is_venv_interpreter(py_path):
    """
    Return True if py_path points to the active venv interpreter.
    """
    venv_root = os.environ.get("VIRTUAL_ENV")
    if not venv_root:
        return False

    venv_root = os.path.realpath(venv_root)
    candidate = os.path.realpath(py_path)
    candidate_raw = os.path.abspath(py_path)

    def _within(root, path):
        try:
            return os.path.commonpath([root, path]) == root
        except ValueError:
            return False

    return _within(venv_root, candidate) or _within(venv_root, candidate_raw)



def remove_env():
    """
    Remove interpreters that belong to the active virtual environment.
    """
    if "VIRTUAL_ENV" not in os.environ:
        return

    with open(DB_FILE, "r", encoding="utf-8") as f:
        rows = list(csv.DictReader(f, delimiter=","))

    with open(DB_FILE, "w", newline="", encoding="utf-8") as f:
        fields = ["PYTHON_VERSION", "PYTHON_PATH"]
        writer = csv.DictWriter(
            f,
            delimiter=",",
            quoting=csv.QUOTE_ALL,
            fieldnames=fields
        )
        writer.writeheader()
        for row in rows:
            python_path = row.get("PYTHON_PATH", "")
            if python_path and _is_venv_interpreter(python_path):
                continue
            writer.writerow({
                "PYTHON_VERSION": row.get("PYTHON_VERSION", ""),
                "PYTHON_PATH": python_path,
            })








#]===========================================================================[#
#] GET VENVS [#==============================================================[#
#]===========================================================================[#

@dataclass
class VenvInfo:
    """_"""
    venv_name: str
    venv_version: str
    site_packages: str
    is_installed: str
    venv_comment: str


def get_venvs(path):
    """
    Get the available virtual environments
    from the specified folder.
    """
    # return an emtpty list if directory doesn't exist
    if not os.path.isdir(path):
        return []

    venv_info_list = []

    for i, venv in enumerate(os.listdir(path)):
        # build path to venv directory
        valid_venv = os.path.join(path, venv)

        # only look for dirs
        if not os.path.isdir(valid_venv):
            continue

        # build path to pyvenv.cfg file
        cfg_file = os.path.join(valid_venv, "pyvenv.cfg")
        if not os.path.isfile(cfg_file):
            continue

        # build path to venvipy.cfg file
        venvipy_cfg_file = os.path.join(valid_venv, "venvipy.cfg")

        venv_name = os.path.basename(valid_venv)
        venv_version = get_config(cfg_file, "version")
        site_packages = get_config(cfg_file, "site_packages")
        is_installed = get_config(cfg_file, "installed")
        venv_comment = get_comment(venvipy_cfg_file)

        venv_info = VenvInfo(
            venv_name,
            venv_version,
            site_packages,
            is_installed,
            venv_comment
        )
        venv_info_list.append(venv_info)

    return venv_info_list[::-1]


def get_config(cfg_file, cfg):
    """
    Return the values as string from a `pyvenv.cfg` file.
    Values for `cfg` can be: `version`, `py_path`,
    `site_packages`, `installed`, `comment`.
    """
    with open(cfg_file, "r", encoding="utf-8") as f:
        lines = f.readlines()

    config = {}

    for line in lines:
        if "=" not in line:
            continue

        key, value = line.split("=", 1)
        config[key.strip().lower()] = value.strip()


    def normalized_version(value):
        numbers = re.findall(r"\d+", value)

        if len(numbers) >= 3:
            return ".".join(numbers[:3])

        if len(numbers) >= 2:
            return ".".join(numbers[:2])

        return value.strip()


    def major_minor(value):
        numbers = re.findall(r"\d+", value)
        if len(numbers) >= 2:
            return f"{numbers[0]}.{numbers[1]}"
        return value.strip()


    raw_version = config.get("version") or config.get("version_info", "")
    version_value = normalized_version(raw_version) if raw_version else "N/A"
    version_str = to_version(version_value) if raw_version else "N/A"

    home = config.get("home", "")
    base_executable = config.get("base-executable", "")
    version_suffix = major_minor(version_value) if raw_version else ""
    binary_path = base_executable or to_path(home, version_suffix)

    site_packages = config.get("include-system-site-packages", "N/A")

    if cfg == "version":
        return version_str

    if cfg == "py_path":
        return binary_path

    if cfg == "site_packages":
        if site_packages == "true":
            return "global"
        if site_packages == "false":
            return "isolated"
        return "N/A"

    if cfg == "installed":
        ensure_dbfile()
        with open(DB_FILE, newline="", encoding="utf-8") as cf:
            reader = csv.DictReader(cf, delimiter=",")
            for info in reader:
                if binary_path == info["PYTHON_PATH"]:
                    return "yes"
            return "no"

    return "N/A"


def get_active_dir_str():
    """Return path to selected directory.
    """
    ensure_active_dir()
    with open(ACTIVE_DIR, "r", encoding="utf-8") as f:
        selected_dir = f.read()
        return selected_dir
    return ""


def get_selected_dir():
    """
    Get the selected directory path from `selected-dir`
    file. Return `get_venvs()`.
    """
    selected_dir = get_active_dir_str()
    return get_venvs(selected_dir)


def get_comment(cfg_file):
    """Get the comment string from `venvipy_cfg` file.
    """
    if os.path.exists(cfg_file):
        with open(cfg_file, "r", encoding="utf-8") as f:
            venv_comment = f.read()

        return venv_comment
    return ""



#]===========================================================================[#
#] GET INFOS FROM PYTHON PACKAGE INDEX [#====================================[#
#]===========================================================================[#

@dataclass
class PackageInfo:
    """
    Info about a PyPI package.
    """
    pkg_name: str
    pkg_version: str
    pkg_info_2: str
    pkg_summary: str



def ensure_pypi_db() -> None:
    CFG_DIR.mkdir(parents=True, exist_ok=True)
    with sqlite3.connect(PACKAGE_DB_PATH) as con:
        con.execute("PRAGMA journal_mode=WAL;")
        con.execute("CREATE TABLE IF NOT EXISTS projects (name TEXT PRIMARY KEY)")
        con.execute("CREATE TABLE IF NOT EXISTS meta (key TEXT PRIMARY KEY, value TEXT)")
        con.execute("CREATE INDEX IF NOT EXISTS idx_projects_name_nocase ON projects(name COLLATE NOCASE)")

        # cache table: version/author/summary + timestamp
        # if older schema exists -> drop & recreate once (simple, no ALTER-mess)
        cols = {r[1] for r in con.execute("PRAGMA table_info(pkg_meta)").fetchall()}
        if cols and not {"name", "version", "info_2", "summary", "fetched_at"}.issubset(cols):
            con.execute("DROP TABLE IF EXISTS pkg_meta")

        con.execute("""
            CREATE TABLE IF NOT EXISTS pkg_meta (
                name TEXT PRIMARY KEY,
                version TEXT NOT NULL DEFAULT '',
                info_2  TEXT NOT NULL DEFAULT '',
                summary TEXT NOT NULL DEFAULT '',
                fetched_at INTEGER NOT NULL DEFAULT 0
            )
        """)

def update_pypi_index(force: bool = False, timeout: int = 60) -> bool:
    """
    Returns True if DB was updated, False if unchanged.
    """
    ensure_pypi_db()

    with sqlite3.connect(PACKAGE_DB_PATH) as con:
        cur = con.cursor()
        cur.execute("SELECT value FROM meta WHERE key='last_serial'")
        row = cur.fetchone()
        stored_serial = row[0] if row else None

    with requests.Session() as s:
        r = s.get(
            PYPI_SIMPLE_URL,
            headers={
                "Accept": "application/vnd.pypi.simple.v1+json, application/json;q=0.9, text/html;q=0.1",
                "User-Agent": "venvipy",
            },
            timeout=timeout,
        )
        r.raise_for_status()

        last_serial = r.headers.get("X-PyPI-Last-Serial")

        if (not force) and last_serial and stored_serial == last_serial:
            return False

        ctype = (r.headers.get("content-type") or "").lower()
        if "json" in ctype:
            data = r.json()
            projects = [p["name"] for p in data.get("projects", []) if isinstance(p, dict) and p.get("name")]
        else:
            soup = BeautifulSoup(r.text, "html.parser")
            projects = [a.get_text(strip=True) for a in soup.select("a") if a.get_text(strip=True)]

    with sqlite3.connect(PACKAGE_DB_PATH) as con:
        con.execute("BEGIN")
        con.execute("DELETE FROM projects")
        con.executemany("INSERT INTO projects(name) VALUES (?)", [(n,) for n in projects])
        con.execute(
            """
            INSERT INTO meta(key, value) VALUES ('last_serial', ?)
            ON CONFLICT(key) DO UPDATE SET value=excluded.value
            """,
            (last_serial or "",),
        )
        con.commit()

    return True


def _get_db_names(name: str, following: int) -> List[str]:
    """
    Returns anchor match + `following` subsequent package names (alphabetical).
    If nothing found, returns [].
    """
    if not name or following < 0:
        return []

    def q1(cur, sql, params) -> Optional[str]:
        cur.execute(sql, params)
        row = cur.fetchone()
        return row[0] if row else None

    with sqlite3.connect(PACKAGE_DB_PATH) as con:
        cur = con.cursor()

        # 1) exact (case-insensitive)
        anchor = q1(
            cur,
            f"SELECT {DB_COL} FROM {DB_TABLE} WHERE lower({DB_COL}) = lower(?) LIMIT 1",
            (name,),
        )

        # 2) prefix (case-insensitive)
        if anchor is None:
            anchor = q1(
                cur,
                f"""
                SELECT {DB_COL}
                FROM {DB_TABLE}
                WHERE {DB_COL} LIKE ? ESCAPE '\\' COLLATE NOCASE
                ORDER BY {DB_COL} COLLATE NOCASE
                LIMIT 1
                """,
                (f"{name}%",),
            )

        # 3) contains (case-insensitive), prefer earlier occurrence + shorter names
        if anchor is None:
            anchor = q1(
                cur,
                f"""
                SELECT {DB_COL}
                FROM {DB_TABLE}
                WHERE {DB_COL} LIKE ? COLLATE NOCASE
                ORDER BY instr(lower({DB_COL}), lower(?)) ASC, length({DB_COL}) ASC, {DB_COL} COLLATE NOCASE
                LIMIT 1
                """,
                (f"%{name}%", name),
            )

        # 4) fallback: next alphabetical position
        if anchor is None:
            anchor = q1(
                cur,
                f"""
                SELECT {DB_COL}
                FROM {DB_TABLE}
                WHERE {DB_COL} >= ? COLLATE NOCASE
                ORDER BY {DB_COL} COLLATE NOCASE
                LIMIT 1
                """,
                (name,),
            )

        if anchor is None:
            return []

        # anchor + following subsequent
        limit = following + 1
        cur.execute(
            f"""
            SELECT {DB_COL}
            FROM {DB_TABLE}
            WHERE {DB_COL} >= ? COLLATE NOCASE
            ORDER BY {DB_COL} COLLATE NOCASE
            LIMIT ?
            """,
            (anchor, limit),
        )
        return [r[0] for r in cur.fetchall()]


def get_db_name(name: str, following: int) -> str:
    """
    Returns a newline-separated string of suggested package names.
    If nothing found, returns "".
    """
    names = _get_db_names(name, following)
    return "\n".join(names) if names else ""


def get_pkg_info_2(name: str, ttl_sec: int = 24 * 3600) -> str:
    """
    Return author (pkg_info_2) for `name`,
    cached in ~/.venvipy/pypi_index.sqlite3.
    """
    if not name:
        return ""

    now = int(time.time())

    with sqlite3.connect(PACKAGE_DB_PATH) as con:
        con.execute("""
            CREATE TABLE IF NOT EXISTS pkg_meta (
                name TEXT PRIMARY KEY,
                version TEXT NOT NULL DEFAULT '',
                info_2  TEXT NOT NULL DEFAULT '',
                summary TEXT NOT NULL DEFAULT '',
                fetched_at INTEGER NOT NULL
            )
        """)

        row = con.execute(
            "SELECT info_2, fetched_at FROM pkg_meta WHERE name = ?",
            (name,),
        ).fetchone()

        if row:
            cached_info_2, fetched_at = row
            if cached_info_2 and (now - int(fetched_at) < ttl_sec):
                return cached_info_2

    # fetch fresh
    try:
        r = requests.get(PYPI_JSON_URL.format(name=name), timeout=10)
        if r.status_code == 404:
            return ""
        r.raise_for_status()
        info = r.json().get("info", {}) or {}
        author = (info.get("author") or "").strip()
        author_email = (info.get("author_email") or "").strip()

        # keep it short but useful
        info_2 = author
        if author_email and author_email not in info_2:
            info_2 = f"{author} <{author_email}>".strip() if author else author_email
    except Exception:
        if row:
            return row[0] or ""
        return ""

    with sqlite3.connect(PACKAGE_DB_PATH) as con:
        con.execute(
            """
            INSERT INTO pkg_meta(name, info_2, fetched_at)
            VALUES (?, ?, ?)
            ON CONFLICT(name) DO UPDATE SET
                info_2 = excluded.info_2,
                fetched_at = excluded.fetched_at
            """,
            (name, info_2, now),
        )

    return info_2


def get_pkg_summary(name: str, ttl_sec: int = 24 * 3600) -> str:
    """
    Return short description (PyPI 'summary')
    for `name`, cached in pkg_meta.summary.
    """
    if not name:
        return ""

    now = int(time.time())

    with sqlite3.connect(PACKAGE_DB_PATH) as con:
        row = con.execute(
            "SELECT summary, fetched_at FROM pkg_meta WHERE name = ?",
            (name,),
        ).fetchone()

    if row:
        cached_summary, fetched_at = row
        if cached_summary and (now - int(fetched_at) < ttl_sec):
            return cached_summary
    else:
        cached_summary = ""

    try:
        r = requests.get(PYPI_JSON_URL.format(name=name), timeout=10)
        if r.status_code == 404:
            return ""
        r.raise_for_status()
        info = r.json().get("info", {}) or {}

        summary = (info.get("summary") or "").strip()
        if not summary:
            # fallback: derive a short line from long description
            desc = (info.get("description") or "").strip()
            summary = re.sub(r"\s+", " ", desc).strip()[:160] if desc else ""
    except Exception:
        return cached_summary or ""

    with sqlite3.connect(PACKAGE_DB_PATH) as con:
        con.execute(
            """
            INSERT INTO pkg_meta(name, summary, fetched_at)
            VALUES (?, ?, ?)
            ON CONFLICT(name) DO UPDATE SET
                summary = excluded.summary,
                fetched_at = excluded.fetched_at
            """,
            (name, summary, now),
        )

    return summary


def get_pkg_version(name: str, ttl_sec: int = 24 * 3600) -> str:
    """
    Return latest version for `name`, cached
    in ~/.venvipy/pypi_index.sqlite3.
    """
    if not name:
        return ""

    now = int(time.time())

    with sqlite3.connect(PACKAGE_DB_PATH) as con:
        con.execute("""
            CREATE TABLE IF NOT EXISTS pkg_meta (
                name TEXT PRIMARY KEY,
                version TEXT NOT NULL,
                fetched_at INTEGER NOT NULL
            )
        """)

        row = con.execute(
            "SELECT version, fetched_at FROM pkg_meta WHERE name = ?",
            (name,),
        ).fetchone()

        if row:
            cached_version, fetched_at = row
            if cached_version and (now - int(fetched_at) < ttl_sec):
                return cached_version

    # fetch fresh
    try:
        r = requests.get(PYPI_JSON_URL.format(name=name), timeout=10)
        if r.status_code == 404:
            return ""
        r.raise_for_status()
        version = (r.json().get("info", {}) or {}).get("version", "") or ""
    except Exception:
        # fallback to cached if available
        if row:
            return row[0] or ""
        return ""

    with sqlite3.connect(PACKAGE_DB_PATH) as con:
        con.execute(
            """
            INSERT INTO pkg_meta(name, version, fetched_at)
            VALUES (?, ?, ?)
            ON CONFLICT(name) DO UPDATE SET
                version = excluded.version,
                fetched_at = excluded.fetched_at
            """,
            (name, version, now),
        )

    return version


def get_package_infos(pkg: str) -> list[PackageInfo]:
    """Get package infos from PyPI (pkg_name from DB for now).
    """
    package_info_list: list[PackageInfo] = []

    # e.g. show 15 suggestions; adjust as needed
    for pkg_name in _get_db_names(pkg, following=15):
        pkg_version = get_pkg_version(pkg_name)
        pkg_info_2 = get_pkg_info_2(pkg_name)
        pkg_summary = get_pkg_summary(pkg_name)

        pkg_info = PackageInfo(
            pkg_name,
            pkg_version,
            pkg_info_2,
            pkg_summary
        )
        package_info_list.append(pkg_info)

    return package_info_list[::-1]


def get_installed_packages(venv_location, venv_name) -> list:
    """Get infos about installed packages.
    """
    platform = get_platform()
    venv_path = Path(venv_location) / venv_name
    site_packages_dir = platform.site_packages_path(venv_path)

    # get list of installed packages
    package_info_list = []
    if not site_packages_dir.exists():
        return package_info_list
    site_packages = os.listdir(site_packages_dir)

    for pkg in site_packages:
        if ".dist-info" in pkg:
            meta_file = site_packages_dir / pkg / "METADATA"

            try:
                with open(meta_file, "r", encoding="utf-8") as f:
                    meta_data = f.readlines()
            except FileNotFoundError:
                logger.debug(f"File '{meta_file}' not found.")
                continue

            pkg_name = ""
            pkg_version = ""
            pkg_info_2 = ""
            pkg_summary = ""

            # search for each str
            for line in meta_data:
                if "Name: " in line:
                    pkg_name = line[5:].strip()

                if "Version: " in line:
                    pkg_version = line[8:].strip()

                if "Author: " in line:
                    pkg_info_2 = line[7:].strip()

                if "Summary: " in line:
                    pkg_summary = line[8:].strip()

            if pkg_name:
                pkg_info = PackageInfo(
                    pkg_name,
                    pkg_version,
                    pkg_info_2,
                    pkg_summary
                )
                package_info_list.append(pkg_info)

    return package_info_list[::-1]






if __name__ == "__main__":

    update_pypi_index(True)
    
    print(get_package_infos("venvipy"))



#    VenviPy - A Virtual Environment Manager for Python.
#    Copyright (C) 2021 - Youssef Serestou - sinusphi.sq@gmail.com
#
#    This program is free software: you can redistribute it and/or modify
#    it under the terms of the GNU General Public License as published by
#    the Free Software Foundation, either version 3 of the License or any
#    later version.
#
#    This program is distributed in the hope that it will be useful,
#    but WITHOUT ANY WARRANTY; without even the implied warranty of
#    MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
#    GNU General Public License for more details.
#
#    A copy of the GNU General Public License version 3 named LICENSE is
#    in the root directory of VenviPy.
#    If not, see <https://www.gnu.org/licenses/licenses.en.html#GPL>.

# -*- coding: utf-8 -*-
"""
This module contains the tables.
"""
import os
import shlex
import subprocess
import tempfile
from pathlib import Path
import shutil
import logging
from functools import partial

from PyQt6.QtGui import QIcon, QCursor, QAction
from PyQt6.QtCore import pyqtSignal, QThread, QTimer
from PyQt6.QtWidgets import (
    QStyle,
    QTableView,
    QMenu,
    QMessageBox,
    QFileDialog,
    QInputDialog
)

import get_data
import creator
from dialogs import ConsoleDialog, ProgBarDialog
from creator import InstallWorker
from manage_pip import PipManager
from platforms import get_platform


logger = logging.getLogger(__name__)



class BaseTable(QTableView):
    """The base table for all tables.
    """
    def __init__(self, *args, **kwargs):
        super().__init__(*args, **kwargs)

        self.delete_icon = QIcon(
            self.style().standardIcon(QStyle.StandardPixmap.SP_TrashIcon)
        )
        self.info_icon = QIcon(
            self.style().standardIcon(QStyle.StandardPixmap.SP_FileDialogInfoView)
        )

    def get_selected_item(self):
        """
        Return `str` from `name` column of the selected row.
        """
        listed_items = self.selectionModel().selectedRows()
        for index in listed_items:
            selected_item = index.data()
            return selected_item


    def get_comment(self):
        """
        Return `str` from `comment` column of the selected row.
        """
        index = self.currentIndex()
        row_index = self.selectionModel().selectedRows()
        row_comment = index.sibling(row_index[0].row(), 4).data()
        return row_comment



class VenvTable(BaseTable):
    """List the virtual environments found.
    """
    started = pyqtSignal()
    finished = pyqtSignal()
    text_changed = pyqtSignal(str)
    refresh = pyqtSignal()
    start_installer = pyqtSignal()
    start_pkg_manager = pyqtSignal()


    def __init__(self, *args, **kwargs):
        super().__init__(*args, **kwargs)

        self.drive_icon = QIcon(
            self.style().standardIcon(QStyle.StandardPixmap.SP_DriveHDIcon)
        )
        self.delete_icon = QIcon(
            self.style().standardIcon(QStyle.StandardPixmap.SP_TrashIcon)
        )
        self.save_icon = QIcon(
            self.style().standardIcon(QStyle.StandardPixmap.SP_DialogSaveButton)
        )
        self.folder_icon = QIcon(
            self.style().standardIcon(QStyle.StandardPixmap.SP_DirOpenIcon)
        )
        self.info_icon = QIcon(
            self.style().standardIcon(QStyle.StandardPixmap.SP_FileDialogInfoView)
        )
        self.reload_icon = QIcon(
            self.style().standardIcon(QStyle.StandardPixmap.SP_BrowserReload)
        )
        self.desk_icon = QIcon(
            self.style().standardIcon(QStyle.StandardPixmap.SP_DesktopIcon)
        )
        self.computer_icon = QIcon(
            self.style().standardIcon(QStyle.StandardPixmap.SP_ComputerIcon)
        )

        self.progress_bar = ProgBarDialog()
        self.console = ConsoleDialog()
        self.thread = QThread(self)
        self.m_install_worker = InstallWorker()

        # thread
        self.thread.start()
        self.m_install_worker.moveToThread(self.thread)

        self.m_install_worker.started.connect(self.console.exec)
        self.m_install_worker.text_changed.connect(
            self.console.update_status
        )

        # perform a proper stop using quit() and wait()
        self.thread.finished.connect(self.thread.quit)
        self.thread.finished.connect(self.thread.wait)


    def contextMenuEvent(self, event):

        idx = self.indexAt(event.pos())
        if not idx.isValid():
            return

        self.selectRow(idx.row())

        #]===================================================================[#
        #] ACTIONS [#========================================================[#
        #]===================================================================[#

        upgrade_pip_action = QAction(
            self.reload_icon,
            "&Upgrade Pip to latest",
            self,
            statusTip="Upgrade Pip to the latest version"
        )
        upgrade_pip_action.triggered.connect(
            lambda: self.upgrade_pip(event, pkg="pip")
        )

        install_wheel_action = QAction(
            self.desk_icon,
            "Install &Wheel",
            self,
            statusTip="Install wheel package"
        )
        install_wheel_action.triggered.connect(
            lambda: self.install_wheel(event, pkg="wheel")
        )

        manage_packages_action = QAction(
            "&Manage installed packages",
            self,
            statusTip="Manage installed packages"
        )
        manage_packages_action.triggered.connect(
            lambda: self.manage_packages(event)
        )

        install_packages_action = QAction(
            "Install &packages from PyPI",
            self,
            statusTip="Install packages from PyPI"
        )
        install_packages_action.triggered.connect(
            lambda: self.install_pypi_packages(event)
        )

        install_requires_action = QAction(
            "Install from a &requirements file",
            self,
            statusTip="Install packages from requirements"
        )
        install_requires_action.triggered.connect(
            lambda: self.install_requires(event)
        )

        install_local_action = QAction(
            "Install from a &local project directory",
            self,
            statusTip="Install from a local stored project directory"
        )
        install_local_action.triggered.connect(
            lambda: self.install_local(event)
        )

        install_vcs_action = QAction(
            "Install from a &VCS project url",
            self,
            statusTip="Install from a VCS url"
        )
        install_vcs_action.triggered.connect(
            lambda: self.install_vcs(event)
        )

        save_requires_action = QAction(
            self.save_icon,
            "Save &requirements",
            self,
            statusTip="Generate a requirements file"
        )
        save_requires_action.triggered.connect(
            lambda: self.save_requires(event)
        )

        comment_add_action = QAction(
            self.info_icon,
            "&Add or modify description",
            self,
            statusTip="Add or modify description"
        )
        comment_add_action.triggered.connect(
            lambda: self.comment_add(event)
        )

        comment_remove_action = QAction(
            self.info_icon,
            "&Remove description",
            self,
            statusTip="Remove description"
        )
        comment_remove_action.triggered.connect(
            lambda: self.comment_remove(event)
        )

        list_packages_action = QAction(
            "&List installed packages",
            self,
            statusTip="List installed packages"
        )
        list_packages_action.triggered.connect(
            lambda: self.list_packages(event, style=1)
        )

        list_freeze_action = QAction(
            "Show &freeze output",
            self,
            statusTip="List the output of 'pip freeze'"
        )
        list_freeze_action.triggered.connect(
            lambda: self.freeze_packages(event, style=2)
        )

        list_deptree_action = QAction(
            "Display &dependency tree",
            self,
            statusTip="List dependencies with 'pipdeptree'"
        )
        list_deptree_action.triggered.connect(
            lambda: self.deptree_packages(event, style=3)
        )

        open_venv_dir_action = QAction(
            self.folder_icon,
            "&Open containing folder",
            self,
            statusTip="Open the folder containing the virtual environment"
        )
        open_venv_dir_action.triggered.connect(
            lambda: self.open_venv_dir(event)
        )

        open_terminal_action = QAction(
            self.desk_icon,
            "Open &terminal (activated)",
            self,
            statusTip="Open a terminal with this environment activated"
        )
        open_terminal_action.triggered.connect(lambda: self.open_terminal(event))

        delete_venv_action = QAction(
            self.delete_icon,
            "Delete &environment",
            self,
            statusTip="Delete environment"
        )
        delete_venv_action.triggered.connect(
            lambda: self.delete_venv(event)
        )

        #]===================================================================[#
        #] MENUS [#==========================================================[#
        #]===================================================================[#

        context_menu = QMenu(self)

        details_sub_menu = QMenu("Det&ails", self, icon=self.info_icon)
        manage_sub_menu  = QMenu("&Manage", self, icon=self.computer_icon)
        comment_sub_menu = QMenu("&Description", self, icon=self.info_icon)

        context_menu.addAction(upgrade_pip_action)
        context_menu.addAction(install_wheel_action)

        context_menu.addMenu(manage_sub_menu)
        manage_sub_menu.addAction(install_packages_action)
        manage_sub_menu.addAction(install_requires_action)
        manage_sub_menu.addAction(install_local_action)
        manage_sub_menu.addAction(install_vcs_action)

        context_menu.addMenu(details_sub_menu)
        details_sub_menu.addAction(list_packages_action)
        details_sub_menu.addAction(list_freeze_action)

        context_menu.addMenu(comment_sub_menu)
        comment_sub_menu.addAction(comment_add_action)
        comment_sub_menu.addAction(comment_remove_action)

        context_menu.addAction(save_requires_action)
        context_menu.addAction(open_venv_dir_action)
        context_menu.addAction(open_terminal_action)
        context_menu.addAction(delete_venv_action)

        context_menu.exec(event.globalPos())


    def valid_version(self, venv_path):
        """Test wether the Python version required is installed.
        """
        cfg_file = os.path.join(venv_path, "pyvenv.cfg")
        is_installed = get_data.get_config(cfg_file, "installed")
        version = get_data.get_config(cfg_file, "version")
        py_path = get_data.get_config(cfg_file, "py_path")
        msg_txt = (
            f"This environment requires {version} \n"
            f"from {py_path} which is \nnot installed.\n"
        )

        if is_installed == "no":
            msg_box = QMessageBox(
                QMessageBox.Icon.Critical,
                "Error",
                msg_txt,
                QMessageBox.StandardButton.Ok,
                self
            )
            msg_box.exec()
            return False
        return True


    def venv_exists(self, path):
        """
        Test wether the directory of the selected environment exists.
        """
        if os.path.exists(path):
            return True

        msg_box = QMessageBox(
            QMessageBox.Icon.Critical,
            "Error",
            "Selected environment could not be found.",
            QMessageBox.StandardButton.Ok,
            self
        )
        msg_box.exec()

        # refresh venv table
        self.refresh.emit()
        return False


    def has_pip(self, venv_dir, venv_name):
        platform = get_platform()
        venv_path = Path(venv_dir) / venv_name

        if not self.venv_exists(venv_path):
            return False

        if not self.valid_version(venv_path):
            return False

        venv_python = platform.venv_python_path(venv_path)
        if not venv_python.exists():
            QMessageBox.information(self, "Info", "Python binary not found in this environment.")
            return False

        try:
            result = subprocess.run(
                [str(venv_python), "-m", "pip", "--version"],
                capture_output=True,
                check=False,          # <-- wichtig
                text=True,
                encoding="utf-8",
                errors="replace",
            )

        except OSError as e:
            QMessageBox.information(self, "Info", f"Could not run Python:\n{e}")
            return False

        if result.returncode == 0:
            return True

        details = (result.stderr or result.stdout or "").strip()
        logger.debug(f"pip --version failed (rc={result.returncode}):\n{details}")
        QMessageBox.information(self, "Info", "This environment has no Pip installed.")
        return False


    def upgrade_pip(self, event, pkg):
        """Run `pip install --upgrade pip` command.
        """
        self.install_sys_pkg(pkg)


    def install_wheel(self, event, pkg):
        """Run `pip install --upgrade wheel` command.
        """
        self.install_sys_pkg(pkg)


    def install_sys_pkg(self, pkg):
        """Run `pip install --upgrade <pkg>` command.
        """
        active_dir = get_data.get_active_dir_str()
        venv = self.get_selected_item()

        if self.has_pip(active_dir, venv):
            self.console.setWindowTitle(f"Installing latest {pkg}")
            logger.debug(f"Installing latest version of {pkg}...")

            self.manager = PipManager(active_dir, venv)
            self.manager.started.connect(self.console.exec)
            self.manager.text_changed.connect(self.console.update_status)

            self.manager.run_pip(creator.cmds[0], [creator.opts[0], f"{pkg}"])

            # clear the content on window close
            if self.console.close:
                self.console.console_window.clear()



    def manage_packages(self, event):
        """Manage installed packages of the selected environment.
        """
        active_dir = get_data.get_active_dir_str()
        venv = self.get_selected_item()
        venv_path = Path(active_dir) / venv

        if self.has_pip(active_dir, venv):
            with open(get_data.ACTIVE_VENV, "w", encoding="utf-8") as f:
                f.write(str(venv_path))

            self.start_pkg_manager.emit()


    def install_pypi_packages(self, event):
        """
        Install packages from [PyPI](https://pypi.org)
        into the selected environment.
        """
        active_dir = get_data.get_active_dir_str()
        venv = self.get_selected_item()
        venv_path = Path(active_dir) / venv

        if self.has_pip(active_dir, venv):
            with open(get_data.ACTIVE_VENV, "w", encoding="utf-8") as f:
                f.write(str(venv_path))

            self.start_installer.emit()


    def install_requires(self, event):
        """
        Install packages from a requirements file into the
        selected environment.
        """
        active_dir = get_data.get_active_dir_str()
        venv = self.get_selected_item()
        platform = get_platform()
        venv_bin = platform.venv_python_path(Path(active_dir) / venv)

        if self.has_pip(active_dir, venv):
            file_name = QFileDialog.getOpenFileName(
                self,
                "Specify a requirements file"
            )
            file_path = file_name[0]

            if file_path != "":
                creator.fix_requirements(file_path)
                logger.debug("Installing from requirements...")
                cmd = [
                    str(venv_bin),
                    "-m",
                    "pip",
                    *shlex.split(creator.cmds[0]),
                    creator.opts[1],
                    file_path,
                ]
                self.console.setWindowTitle(
                    "Installing from requirements file"
                )
                wrapper = partial(self.m_install_worker.run_process, cmd)
                QTimer.singleShot(0, wrapper)

                # clear the content on window close
                if self.console.close:
                    self.console.console_window.clear()


    def install_local(self, event):
        """Install from a local project directory.
        """
        active_dir = get_data.get_active_dir_str()
        venv = self.get_selected_item()

        if self.has_pip(active_dir, venv):
            project_dir = QFileDialog.getExistingDirectory(
                self,
                "Select project directory"
            )
            project_name = os.path.basename(project_dir)

            if project_dir != "":
                if get_data.is_writable(project_dir):
                    self.console.setWindowTitle(f"Installing {project_name}")
                    logger.debug("Installing from local project directory...")

                    self.manager = PipManager(active_dir, venv)
                    self.manager.started.connect(self.console.exec)
                    self.manager.text_changed.connect(self.console.update_status)
                    self.manager.run_pip(
                        creator.cmds[0], [creator.opts[3], project_dir]
                    )

                    # clear the content on window close
                    if self.console.close:
                        self.console.console_window.clear()

                else:
                    msg_txt = (
                        "Filesystem not writable.\n\n"
                        "Unable to create the '.egg-info'      \n"
                        "folder inside the project directory.      \n\n"
                    )
                    QMessageBox.warning(self, "Abort", msg_txt)


    def install_vcs(self, event):
        """Install from VCS.
        """
        active_dir = get_data.get_active_dir_str()
        venv = self.get_selected_item()
        platform = get_platform()
        venv_bin = platform.venv_python_path(Path(active_dir) / venv)

        if self.has_pip(active_dir, venv):
            url, ok = QInputDialog.getText(
                self,
                "Specify project",
                "Enter link to repository:" + " " * 80
            )

            if url != "":
                self.vcs_project_url = url[:-4] if url.endswith(".git") else url
                self.vcs_project_name = os.path.basename(self.vcs_project_url)
                formatted_project_url = (
                    f"git+{self.vcs_project_url}#egg={self.vcs_project_name}"
                )
                cmd = [
                    str(venv_bin),
                    "-m",
                    "pip",
                    *shlex.split(creator.cmds[0]),
                    formatted_project_url,
                ]
                self.console.setWindowTitle(
                    f"Installing {self.vcs_project_name}"
                )
                wrapper = partial(self.m_install_worker.run_process, cmd)
                QTimer.singleShot(0, wrapper)

                # clear the content on window close
                if self.console.close:
                    self.console.console_window.clear()


    def save_requires(self, event):
        """
        Pip freeze a requirements of the selected environment.
        """
        active_dir = get_data.get_active_dir_str()
        venv = self.get_selected_item()
        venv_dir = Path(active_dir) / venv

        if self.has_pip(active_dir, venv):
            save_file = QFileDialog.getSaveFileName(
                self,
                "Save requirements",
                directory=str(venv_dir / "requirements.txt")
            )
            save_path = save_file[0]

            if not save_path:
                return

            platform = get_platform()
            venv_python = platform.venv_python_path(venv_dir)
            try:
                with open(save_path, "w", encoding="utf-8") as f:
                    subprocess.run([
                        str(venv_python),
                        "-m",
                        "pip",
                        "freeze"],
                        stdout=f,
                        check=False,
                    )
            except subprocess.CalledProcessError as e:
                logger.debug(f"Failed to save requirements: {e}")

            logger.debug(f"Saved '{save_path}'")
            message_txt = f"Saved requirements in \n{save_path}"
            QMessageBox.information(self, "Saved", message_txt)

            # comment the 'pkg_resources==0.0.0' entry
            # DEPRECATED: now handled in manage_pip.py
            #creator.fix_requirements(save_path)


    def list_packages(self, event, style):
        """
        Open console dialog and list the installed packages. The argument
        `style` controls which style the output should have: `style=1` for
        `pip list`, `style=2` for `pip freeze` and style=3 for a dependency
        output via `pipdeptree`.
        """
        active_dir = get_data.get_active_dir_str()
        venv = self.get_selected_item()

        if self.has_pip(active_dir, venv):
            self.console.setWindowTitle(f"Packages in: {venv}")

            self.manager = PipManager(active_dir, venv)
            self.manager.started.connect(self.console.exec)
            self.manager.text_changed.connect(self.console.update_status)

            self.manager.run_pip(creator.cmds[style])

            # clear the content on window close
            if self.console.close:
                self.console.console_window.clear()


    def freeze_packages(self, event, style):
        """Print `pip freeze` output to console window.
        """
        self.list_packages(event, style)


    def deptree_packages(self, event, style):
        """
        Test if `pipdeptree` is installed and ask user wether to
        install it if it's not. Then call `self.list_packages()`
        """
        active_dir = get_data.get_active_dir_str()
        venv = self.get_selected_item()
        platform = get_platform()
        venv_python = platform.venv_python_path(Path(active_dir) / venv)
        pipdeptree_result = subprocess.run(
            [str(venv_python), "-m", "pipdeptree", "--version"],
            capture_output=True,
            check=False,
            text=True,
            encoding="utf-8",
            errors="replace",
        )
        has_pipdeptree = pipdeptree_result.returncode == 0
        message_txt = (
            "This requires pipdeptree             \nto be installed.\n\n"
            "Do you want to install it?\n"
        )

        if has_pipdeptree:
            self.list_packages(event, style)
        else:
            if self.has_pip(active_dir, venv):
                msg_box_confirm = QMessageBox.question(
                    self,
                    "Confirm",
                    message_txt,
                    QMessageBox.StandardButton.Yes
                    | QMessageBox.StandardButton.Cancel
                )
                if msg_box_confirm == QMessageBox.StandardButton.Yes:
                    self.progress_bar.setWindowTitle("Installing")
                    self.progress_bar.status_label.setText(
                        "Installing pipdeptree..."
                    )
                    logger.debug("Installing pipdeptree...")

                    self.manager = PipManager(active_dir, venv)
                    self.manager.run_pip(
                        creator.cmds[0], [creator.opts[0], "pipdeptree"]
                    )
                    self.manager.started.connect(self.progress_bar.exec)
                    self.manager.finished.connect(self.progress_bar.close)
                    self.manager.process_stop()
                    self.list_packages(event, style)


    def comment_add(self, event):
        """Add / modify a comment.
        """
        active_dir = get_data.get_active_dir_str()
        venv = self.get_selected_item()
        comment_old = self.get_comment()
        venvipy_cfg = os.path.join(active_dir, venv, "venvipy.cfg")

        if not os.path.exists(venvipy_cfg):
            comment_new, ok = QInputDialog.getText(
                self,
                "Enter description",
                "\nEnter a description:" + " " * 70
            )
            if len(comment_new) > 0:
                # save a new description
                creator.save_comment(venvipy_cfg, comment_new)
                logger.debug("Description saved")

        else:
            comment_new, ok = QInputDialog.getText(
                self,
                "Modify description",
                "\nModify your description:" + " " * 70,
                text=comment_old
            )
            if len(comment_new) > 0:
                # modify an existing description
                creator.save_comment(venvipy_cfg, comment_new)
                logger.debug("Description updated")
            else:
                # keep the old description
                creator.save_comment(venvipy_cfg, comment_old)

        # refresh venv table
        self.refresh.emit()


    def comment_remove(self, event):
        """Remove comment.
        """
        active_dir = get_data.get_active_dir_str()
        venv = self.get_selected_item()
        venvipy_cfg = os.path.join(active_dir, venv, "venvipy.cfg")

        if os.path.exists(venvipy_cfg):
            msg_box_warning = QMessageBox.warning(
                self,
                "Remove comment",
                "Delete this comment.        \n"
                "Are you sure?\n",
                QMessageBox.StandardButton.Yes
                | QMessageBox.StandardButton.Cancel
            )
            if msg_box_warning == QMessageBox.StandardButton.Yes:
                os.remove(venvipy_cfg)
                logger.debug(f"Successfully deleted '{venvipy_cfg}'")
                self.refresh.emit()


    def open_venv_dir(self, event):
        """Open the selected venv directory.
        """
        active_dir = get_data.get_active_dir_str()
        venv = self.get_selected_item()
        venv_dir = Path(active_dir) / venv

        if venv_dir.is_dir():
            platform = get_platform()
            if platform.is_windows():
                os.startfile(str(venv_dir))
            else:
                subprocess.run(["xdg-open", str(venv_dir)], check=False)

 
    def open_terminal(self, event):
        """Open a terminal with the selected environment activated.

        Note: there is no single "default terminal" API on Linux, so this is
        a best-effort launcher for common terminal emulators.
        """
        active_dir = get_data.get_active_dir_str()
        venv = self.get_selected_item()
        venv_dir = Path(active_dir) / venv

        if not self.venv_exists(str(venv_dir)):
            return
        
        if not self.valid_version(str(venv_dir)):
            return

        platform = get_platform()

        if platform.is_windows():
            try:
                platform.open_activated_terminal(
                    venv_dir=venv_dir,
                    workdir=Path.home(),
                    env=env,
                    banner=True,
                )
            except Exception as e:
                QMessageBox.information(
                    self, "Info", f"Failed to open terminal: {e}"
                )
            return

        activate = platform.activate_script_path(venv_dir)

        if not activate.exists():
            QMessageBox.information(
                self, "Info", "Activate script not found in this environment."
            )
            return

        workdir = Path.home()

        # real rcfile -> works reliably from Python
        fd, rc_path = tempfile.mkstemp(prefix="venvipy-", suffix=".bashrc")
        with os.fdopen(fd, "w", encoding="utf-8") as f:
            f.write(f"source {shlex.quote(str(activate))}\n")
            f.write(f"cd {shlex.quote(str(workdir))}\n\n")
            f.write("# quick self-checks\n")
            f.write("echo Active environment: \"$VIRTUAL_ENV\"\n")
            f.write("echo Python interpreter: \"$(command -v python)\"\n")
            f.write("echo ' '\n")
        os.chmod(rc_path, 0o600)

        env = os.environ.copy()

        # remove inherited venv (the one VenviPy was launched with)
        dev_venv = env.pop("VIRTUAL_ENV", None)
        for k in ("VIRTUAL_ENV_PROMPT", "_OLD_VIRTUAL_PATH", "_OLD_VIRTUAL_PYTHONHOME", "PYTHONHOME"):
            env.pop(k, None)

        # optional: conda leftovers
        for k in ("CONDA_PREFIX", "CONDA_DEFAULT_ENV", "CONDA_SHLVL"):
            env.pop(k, None)

        # also remove the dev venv bin prefix from PATH (otherwise python/pip may still resolve to dev)
        if dev_venv:
            dev_bin = str(Path(dev_venv) / platform.venv_bin_dir_name())
            parts = env.get("PATH", "").split(os.pathsep)
            env["PATH"] = os.pathsep.join(
                p for p in parts
                if Path(p).expanduser().resolve() != Path(dev_bin).expanduser().resolve()
            )

        shell = env.get("SHELL", "/bin/bash")
        shell_name = Path(shell).name

        if shell_name == "zsh" and shutil.which("zsh"):
            zd = tempfile.mkdtemp(prefix="venvipy-zsh-")
            zrc = Path(zd) / ".zshrc"

            zrc.write_text(
                (
                    '[[ -f ~/.zshrc ]] && source ~/.zshrc\n'
                    f"source {shlex.quote(str(activate))}\n"
                    f"cd {shlex.quote(str(workdir))}\n"
                    'echo VIRTUAL_ENV="$VIRTUAL_ENV"\n'
                    "which python\n"
                ),
                encoding="utf-8",
            )
            env["ZDOTDIR"] = zd
            shell_cmd = ["zsh", "-i"]

        else:
            # fallback: your existing bash --rcfile method
            shell_cmd = ["bash", "--rcfile", rc_path, "-i"]

        candidates = [
            ["xfce4-terminal", "-x", *shell_cmd],
            ["gnome-terminal", "--", *shell_cmd],
            ["konsole", "-e", *shell_cmd],
            ["mate-terminal", "--", *shell_cmd],
            ["x-terminal-emulator", "-e", *shell_cmd],
            ["qterminal", "-e", *shell_cmd],
            ["terminator", "-x", *shell_cmd],
            ["alacritty", "-e", *shell_cmd],
            ["kitty", *shell_cmd],
            ["xterm", "-e", *shell_cmd],
        ]

        last_err = None

        for cmd in candidates:
            if not shutil.which(cmd[0]):
                continue

            try:
                subprocess.Popen(cmd, start_new_session=True, env=env)
                return

            except Exception as e:
                last_err = e

        if last_err:
            logger.debug(f"Terminal launch failed: {last_err}")

        QMessageBox.information(
            self,
            "Info",
            "No supported terminal emulator found. Install one and try again."
        )


    def delete_venv(self, event):
        """
        Delete the selected virtual environment by clicking
        delete from the context menu in venv table.
        """
        active_dir = get_data.get_active_dir_str()
        venv = self.get_selected_item()
        venv_path = os.path.join(active_dir, venv)

        if self.venv_exists(venv_path):
            msg_box_critical = QMessageBox.critical(
                self,
                "Delete venv",
                f"Delete '{venv}'?           \n"
                "Are you sure? This can't be undone.\n",
                QMessageBox.StandardButton.Yes
                | QMessageBox.StandardButton.Cancel
            )
            if msg_box_critical == QMessageBox.StandardButton.Yes:
                shutil.rmtree(venv_path)
                logger.debug(f"Successfully deleted '{venv_path}'")
                self.refresh.emit()



class InterpreterTable(BaseTable):
    """
    List the Python installs found.
    """
    drop_item = pyqtSignal()

    def __init__(self, *args, **kwargs):
        super().__init__(*args, **kwargs)


    def contextMenuEvent(self, event):
        idx = self.indexAt(event.pos())
        if not idx.isValid():
            return

        self.selectRow(idx.row())

        context_menu = QMenu(self)

        remove_py_action = QAction(
            self.delete_icon,
            "&Remove from list",
            self,
            statusTip="Remove this item from the table"
        )
        remove_py_action.triggered.connect(lambda: QTimer.singleShot(0, lambda: self.remove_python(event)))
        context_menu.addAction(remove_py_action)

        context_menu.exec(event.globalPos())


    def remove_python(self, event):
        """Remove a Python version from the table.
        """
        item = self.get_selected_item()

        msg_box_warning = QMessageBox.warning(
            self,
            "Confirm",
            "Remove this item from list.       \n"
            "Are you sure?\n",
            QMessageBox.StandardButton.Yes | QMessageBox.StandardButton.Cancel
        )
        if msg_box_warning == QMessageBox.StandardButton.Yes:
            with open(get_data.DB_FILE, "r", encoding="utf-8") as f:
                lines = f.readlines()
            with open(get_data.DB_FILE, "w", encoding="utf-8") as f:
                for line in lines:
                    if item not in line:
                        f.write(line)

            logger.debug(f"Removed '{item}' from database")
            self.drop_item.emit()



#    VenviPy - A Virtual Environment Manager for Python.
#    Copyright (C) 2021 - Youssef Serestou - sinusphi.sq@gmail.com
#
#    This program is free software: you can redistribute it and/or modify
#    it under the terms of the GNU General Public License as published by
#    the Free Software Foundation, either version 3 of the License or any
#    later version.
#
#    This program is distributed in the hope that it will be useful,
#    but WITHOUT ANY WARRANTY; without even the implied warranty of
#    MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
#    GNU General Public License for more details.
#
#    A copy of the GNU General Public License version 3 named LICENSE is
#    in the root directory of VenviPy.
#    If not, see <https://www.gnu.org/licenses/licenses.en.html#GPL>.

# -*- coding: utf-8 -*-
"""
This module creates all the stuff requested.
"""
import logging
import shlex
import os
from pathlib import Path
from subprocess import Popen, PIPE, STDOUT, run
from random import randint

from PyQt6.QtCore import QObject, pyqtSignal, pyqtSlot

from manage_pip import PipManager


logger = logging.getLogger(__name__)

# commands / options
cmds = [
    "install --no-cache-dir",  # 0
    "list",  # 1
    "freeze",  # 2
    "pipdeptree",  # 3
]
opts = [
    "--upgrade",  # 0
    "--requirement",  # 1
    "--editable",  # 2
    "--use-feature=in-tree-build",  # 3
]



#]===========================================================================[#
#] CUSTOM WORKER [#==========================================================[#
#]===========================================================================[#

class InstallWorker(QObject):
    """
    This worker performs package installs.
    """
    started = pyqtSignal()
    finished = pyqtSignal()
    text_changed = pyqtSignal(str)

    @pyqtSlot(str)
    def run_process(self, command):
        """Run the process.
        """
        self.started.emit()
        self.install_process(command)
        self.finished.emit()


    def install_process(self, command):
        """Install a package via subprocess.
        """
        os.environ["PYTHONUNBUFFERED"] = "1"
        errors = []

        args = shlex.split(command) if isinstance(command, str) else command

        with Popen(
            args,
            stdout=PIPE,
            stderr=STDOUT,
            encoding="utf-8",
            errors="replace",
        ) as process:
            while process.poll() is None:
                output = process.stdout.readline()

                if output != "":
                    logger.debug(output.strip())
                    if output.lstrip().startswith((
                        "ERROR",
                        "WARNING",
                        "fatal",
                        "remote"
                    )):
                        errors.append(output.strip())
                    else:
                        self.text_changed.emit(output.strip())

            if errors:
                errors.insert(0, " ")
                for line in errors:
                    self.text_changed.emit(line)

            self.text_changed.emit("\n\nPress [ESC] to continue...\n")
            logger.debug(f"Exit code: {process.returncode}")
            return process.returncode



#]===========================================================================[#
#] WORKER (CREATE VIRTUAL ENVIRONMENT) [#====================================[#
#]===========================================================================[#

class CreationWorker(QObject):
    """
    Worker that performs the creation process.
    """
    started = pyqtSignal()
    finished = pyqtSignal()
    updating_pip = pyqtSignal()
    installing_wheel = pyqtSignal()

    @pyqtSlot(tuple)
    def install_venv(self, args):
        """Execute the commands to create the environment.
        """
        self.started.emit()
        logger.debug("Creating virtual environment...")

        py_vers, name, location, with_pip, with_wheel, site_packages = args
        env_dir = Path(location) / name

        create_venv(
            py_vers,
            env_dir,
            with_pip=with_pip,
            system_site_packages=site_packages
        )

        if with_pip and not with_wheel:
            self.manager = PipManager(location, name)
            self.updating_pip.emit()
            self.manager.run_pip(cmds[0], [opts[0], "pip"])
            self.manager.finished.connect(self.finished.emit)
        elif with_pip and with_wheel:
            self.manager = PipManager(location, name)
            self.installing_wheel.emit()
            self.manager.run_pip(cmds[0], [opts[0], "pip", "wheel"])
            self.manager.finished.connect(self.finished.emit)
        else:
            self.finished.emit()


#]===========================================================================[#
#] CREATE A VIRTUAL ENVIRONMENT [#===========================================[#
#]===========================================================================[#

def create_venv(
        py_vers,
        env_dir,
        with_pip=False,
        system_site_packages=False
    ):
    """Create a virtual environment in a directory.
    """
    env_path = Path(env_dir)
    env_path.parent.mkdir(parents=True, exist_ok=True)

    args = [py_vers, "-m", "venv", str(env_path)]
    if not with_pip:
        args.append("--without-pip")
    if system_site_packages:
        args.append("--system-site-packages")

    res = run(
        args,
        stdout=PIPE,
        stderr=STDOUT,
        encoding="utf-8",
        errors="replace",
        text=True,
        check=False,
    )

    return (res.stdout or "").strip()


#]===========================================================================[#
#] REQUIREMENTS [#===========================================================[#
#]===========================================================================[#

def fix_requirements(require_file):
    """
    Check the selected requirements file. If it contains
    a `pkg_resources==0.0.0` entry, then comment this line
    to prevent pip from crashing.
    """
    with open(require_file, "r", encoding="utf-8") as f:
        content = f.readlines()

    new_content = []

    for line in content:
        if line.startswith("pkg_resources==0.0.0"):
            line = line.replace(
                "pkg_resources==0.0.0",
                "#pkg_resources==0.0.0"
            )
            logger.debug(f"Fixed requirements in '{require_file}'")
        else:
            logger.debug(f"Fixing not required for '{require_file}'")

        new_content.append(line)

    with open(require_file, "w", encoding="utf-8") as f:
        f.writelines(new_content)


#]===========================================================================[#
#] DESCRIPTION [#============================================================[#
#]===========================================================================[#

def save_comment(path, comment):
    """Save a description in `venvipy.cfg` file.
    """
    with open(path, "w+", encoding="utf-8") as f:
        f.write(comment)


#]===========================================================================[#
#] GENERATE A RANDOM LINE FROM THE ZEN OF PYTHON [#==========================[#
#]===========================================================================[#

def random_zen_line():
    """Return a random line from the Zen of Python.
    """
    this = [
        "\nBeautiful is better than ugly.",
        "\nExplicit is better than implicit.",
        "\nSimple is better than complex.",
        "\nComplex is better than complicated.",
        "\nFlat is better than nested.",
        "\nSparse is better than dense.",
        "\nReadability counts.",
        "\nSpecial cases aren't special enough\nto break the rules.\nAlthough practicality beats purity.",
        "\nErrors should never pass silently.\nUnless explicitly silenced.",
        "\nIn the face of ambiguity,\nrefuse the temptation to guess.",
        "\nThere should be one\n-- and preferably only one --\nobvious way to do it.\nAlthough that way may not be obvious at first\nunless you're Dutch.",
        "\nNow is better than never.\nAlthough never is often better\nthan *right* now.",
        "\nIf the implementation is hard to explain,\nit's a bad idea.",
        "\nIf the implementation is easy to explain,\nit may be a good idea.",
        "\nNamespaces are one honking great idea\n---\nlet's do more of those!"
    ]
    return this[randint(0, 14)]



if __name__ == "__main__":
    pass



#    VenviPy - A Virtual Environment Manager for Python.
#    Copyright (C) 2021 - Youssef Serestou - sinusphi.sq@gmail.com
#
#    This program is free software: you can redistribute it and/or modify
#    it under the terms of the GNU General Public License as published by
#    the Free Software Foundation, either version 3 of the License or any
#    later version.
#
#    This program is distributed in the hope that it will be useful,
#    but WITHOUT ANY WARRANTY; without even the implied warranty of
#    MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
#    GNU General Public License for more details.
#
#    A copy of the GNU General Public License version 3 named LICENSE is
#    in the root directory of VenviPy.
#    If not, see <https://www.gnu.org/licenses/licenses.en.html#GPL>.



#    VenviPy - A Virtual Environment Manager for Python.
#    Copyright (C) 2021 - Youssef Serestou - sinusphi.sq@gmail.com
#
#    This program is free software: you can redistribute it and/or modify
#    it under the terms of the GNU General Public License as published by
#    the Free Software Foundation, either version 3 of the License or any
#    later version.
#
#    This program is distributed in the hope that it will be useful,
#    but WITHOUT ANY WARRANTY; without even the implied warranty of
#    MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
#    GNU General Public License for more details.
#
#    A copy of the GNU General Public License version 3 named LICENSE is
#    in the root directory of VenviPy.
#    If not, see <https://www.gnu.org/licenses/licenses.en.html#GPL>.

# -*- coding: utf-8 -*-
"""
Themes
"""

# Base palette for the entire project (dark-focused).
COLORS = {
    "bg": "#0f1218",
    "surface": "#151a24",
    "surface_alt": "#1a2130",
    "menu_bar_bg": "#0f1218",
    "menu_bg": "#131a26",
    "panel": "#202938",
    "panel_alt": "#263144",
    "border": "#2c3850",
    "border_strong": "#35435f",
    "text": "#e6e9ef",
    "text_muted": "#a8b0bf",
    "accent": "#4ea3ff",
    "accent_soft": "#1b3556",
    "accent_strong": "#2f7be5",
    "success": "#22c55e",
    "warning": "#f59e0b",
    "danger": "#ef4444",
    "tooltip": "#1f2634",
    "tooltip_text": "#e6e9ef",
    "hover": "#2a3a55",
    "shadow": "#0b0d12",
}

dark = f"""
    QMainWindow {{
        background-color: {COLORS["bg"]};
        border: 1px solid {COLORS["border"]};
        margin: 1px;
        border-radius: 6px;
    }}

    QWidget#centralwidget {{
        background-color: {COLORS["surface"]};
        border-radius: 6px;
    }}

    QMenuBar {{
        background-color: {COLORS["menu_bar_bg"]};
        color: {COLORS["text"]};
        font: 600 13px;
    }}

    QMenuBar::item {{
        background-color: {COLORS["menu_bar_bg"]};
        color: {COLORS["text"]};
    }}


    QMenuBar::item::selected {{
        background-color: {COLORS["panel_alt"]};
    }}

    QWidget#titleBar {{
        background-color: {COLORS["surface_alt"]};
        border-top-left-radius: 6px;
        border-top-right-radius: 6px;
    }}

    QWidget#titleSeparator {{
        background-color: {COLORS["border"]};
    }}

    QLabel#titleLabel {{
        color: {COLORS["text"]};
        font: 600 13px;
    }}

    QToolButton#titleButton {{
        background-color: transparent;
        border: none;
        padding: 4px;
        border-radius: 4px;
    }}

    QToolButton#titleButton:hover {{
        background-color: {COLORS["hover"]};
    }}

    QToolButton#closeButton {{
        background-color: transparent;
        border: none;
        padding: 4px;
        border-radius: 4px;
    }}

    QToolButton#closeButton:hover {{
        background-color: {COLORS["danger"]};
    }}

    QMenu {{
        background-color: {COLORS["menu_bg"]};
        color: {COLORS["text"]};
        border: 1px solid {COLORS["border"]};
    }}

    QMenu::item::selected {{
        background-color: {COLORS["panel_alt"]};
    }}

    QMenu QCheckBox {{
        color: {COLORS["text"]};
        background-color: transparent;
        margin: 0px;
        padding: 4px 12px 5px 4px;
        spacing: 6px;
    }}

    QMenu QCheckBox:hover {{
        background-color: {COLORS["panel_alt"]};
    }}

    QMenu QCheckBox::indicator {{
        width: 10;
        height: 10;
        margin: 0px 6px 0px 0px;
        border: 1px solid {COLORS["border_strong"]};
        border-radius: 2px;
        background-color: {COLORS["surface_alt"]};
    }}

    QMenu QCheckBox::indicator:checked {{
        background-color: {COLORS["accent"]};
        border-color: {COLORS["accent"]};
    }}

    QDialog,
    QMessageBox,
    QInputDialog {{
        background-color: {COLORS["surface"]};
        color: {COLORS["text"]};
    }}

    QDialog QLabel,
    QMessageBox QLabel,
    QInputDialog QLabel {{
        color: {COLORS["text"]};
    }}

    QDialog QLineEdit,
    QMessageBox QLineEdit,
    QInputDialog QLineEdit {{
        background-color: {COLORS["surface_alt"]};
        color: {COLORS["text"]};
        border: 1px solid {COLORS["border"]};
        border-radius: 6px;
        padding: 6px 8px;
    }}

    QDialog QLineEdit:focus,
    QMessageBox QLineEdit:focus,
    QInputDialog QLineEdit:focus {{
        border-color: {COLORS["accent"]};
    }}

    QDialogButtonBox QPushButton {{
        background-color: {COLORS["panel"]};
        border: 1px solid {COLORS["border"]};
        border-radius: 8px;
        padding: 8px 12px;
        color: {COLORS["text"]};
        font: 600 13px;
    }}

    QDialogButtonBox QPushButton:hover {{
        background-color: {COLORS["accent_strong"]};
        border-color: {COLORS["accent"]};
    }}

    QToolTip {{
        background-color: {COLORS["tooltip"]};
        border: 1px solid {COLORS["border"]};
        color: {COLORS["tooltip_text"]};
        padding: 4px 6px;
        opacity: 220;
    }}
    
    QListWidget {{
        background-color: {COLORS["surface_alt"]};
        color: {COLORS["text"]};
        border-radius: 6px;
        border: 1px solid {COLORS["border"]};
        padding: 6px;
    }}

    QListWidget::item {{
        border-radius: 6px;
        padding: 12px;
    }}

    QListWidget::item:hover {{
        background-color: {COLORS["hover"]};
    }}

    QListWidget::item:selected {{
        background-color: {COLORS["accent_soft"]};
        color: {COLORS["text"]};
    }}

    QTabWidget {{
        background-color: {COLORS["surface"]};
        border-radius: 6px;
    }}

    QTabWidget::pane {{
        background-color: {COLORS["surface"]};
        border: 1px solid {COLORS["border"]};
        border-radius: 6px;
    }}

    QTabBar::tab {{
        background: {COLORS["panel"]};
        padding: 8px 12px;
        color: {COLORS["text_muted"]};
        border: 1px solid {COLORS["border"]};
        border-bottom: none;
        border-top-left-radius: 6px;
        border-top-right-radius: 6px;
        margin-right: 2px;
    }}

    QTabBar::tab:selected {{
        background-color: {COLORS["surface_alt"]};
        color: {COLORS["text"]};
        border-color: {COLORS["border_strong"]};
    }}

    QPushButton {{
        background-color: {COLORS["panel"]};
        border: 1px solid {COLORS["border"]};
        border-radius: 8px;
        padding: 8px 12px;
        color: {COLORS["text"]};
        font: 600 13px;
    }}

    QPushButton:hover {{
        background-color: {COLORS["accent_strong"]};
        border-color: {COLORS["accent"]};
    }}

    QPushButton:pressed {{
        background-color: {COLORS["panel_alt"]};
    }}

    QTableView {{
        background-color: {COLORS["surface_alt"]};
        alternate-background-color: {COLORS["surface"]};
        color: {COLORS["text"]};
        border: 1px solid {COLORS["border"]};
        gridline-color: {COLORS["border"]};
        border-radius: 6px;
    }}

    QHeaderView::section {{
        background-color: {COLORS["panel"]};
        color: {COLORS["text"]};
        padding: 6px;
        border: 1px solid {COLORS["border"]};
    }}

    QTableView::item {{
        padding: 8px;
        color: {COLORS["text"]};
    }}

    QTableView::item:selected {{
        background-color: {COLORS["accent_soft"]};
        color: {COLORS["text"]};
    }}

    QScrollBar:vertical {{
        background-color: {COLORS["surface"]};
        width: 12px;
        margin: 0px;
    }}

    QScrollBar::handle:vertical {{
        background-color: {COLORS["panel_alt"]};
        min-height: 20px;
        border-radius: 6px;
        border: 1px solid {COLORS["border"]};
    }}

    QScrollBar:horizontal {{
        background-color: {COLORS["surface"]};
        height: 12px;
        margin: 0px;
    }}

    QScrollBar::handle:horizontal {{
        background-color: {COLORS["panel_alt"]};
        min-width: 20px;
        border-radius: 6px;
        border: 1px solid {COLORS["border"]};
    }}

    QScrollBar::add-line:horizontal, QScrollBar::sub-line:horizontal {{
        width: 0px;
    }}

    QScrollBar::add-line:vertical, QScrollBar::sub-line:vertical {{
        height: 0px;
    }}

    QTableView QTableCornerButton::section {{
        background-color: {COLORS["panel"]};
        border: 1px solid {COLORS["border"]};
    }}

    QTableView QHeaderView {{
        background-color: {COLORS["panel"]};
    }}
"""


light = f"""
    QMainWindow {{
        background-color: {COLORS["bg"]};
        border: 1px solid {COLORS["border"]};
        margin: 1px;
        border-radius: 6px;
    }}

    QWidget#centralwidget {{
        background-color: {COLORS["surface"]};
        border-radius: 6px;
    }}

    QListWidget {{
        background-color: {COLORS["surface_alt"]};
        color: {COLORS["text"]};
        border: 1px solid {COLORS["border"]};
        padding: 6px;
        border-radius: 6px;
    }}

    QListWidget::item {{
        padding: 12px;
    }}

    QListWidget::item:hover {{
        background-color: {COLORS["hover"]};
    }}

    QListWidget::item:selected {{
        background-color: {COLORS["accent_soft"]};
        color: {COLORS["text"]};
    }}

    QTabWidget::pane {{
        background-color: {COLORS["surface"]};
        border-top: 1px solid {COLORS["border"]};
        border-radius: 6px;
    }}

    QTabBar::tab {{
        background: {COLORS["panel"]};
        padding: 8px 12px;
        color: {COLORS["text_muted"]};
        border: 1px solid {COLORS["border"]};
        border-bottom: none;
        border-top-left-radius: 6px;
        border-top-right-radius: 6px;
    }}

    QTabBar::tab:selected {{
        background-color: {COLORS["surface_alt"]};
        color: {COLORS["text"]};
        border-color: {COLORS["border_strong"]};
    }}
                    
    QPushButton {{
        background-color: {COLORS["panel"]};
        border: 1px solid {COLORS["border"]};
        border-radius: 8px;
        padding: 8px 12px;
        color: {COLORS["text"]};
        font: 600 13px;
    }}
                    
    QPushButton:hover {{
        background-color: {COLORS["accent_strong"]};
        border-color: {COLORS["accent"]};
    }}
                    
    QTableView {{
        background-color: {COLORS["surface_alt"]};
        alternate-background-color: {COLORS["surface"]};
        color: {COLORS["text"]};
        border: 1px solid {COLORS["border"]};
        gridline-color: {COLORS["border"]};
        border-radius: 6px;
    }}

    QHeaderView::section {{
        background-color: {COLORS["panel"]};
        color: {COLORS["text"]};
        padding: 6px;
        border: 1px solid {COLORS["border"]};
    }}

    QTableView::item {{
        padding: 8px;
        color: {COLORS["text"]};
    }}

    QTableView::item:selected {{
        background-color: {COLORS["accent_soft"]};
        color: {COLORS["text"]};
    }}

    QScrollBar:vertical {{
        background-color: {COLORS["surface"]};
        width: 12px;
        margin: 0px;
    }}

    QScrollBar::handle:vertical {{
        background-color: {COLORS["panel_alt"]};
        min-height: 20px;
        border-radius: 6px;
        border: 1px solid {COLORS["border"]};
    }}

    QScrollBar:horizontal {{
        background-color: {COLORS["surface"]};
        height: 12px;
        margin: 0px;
    }}

    QScrollBar::handle:horizontal {{
        background-color: {COLORS["panel_alt"]};
        min-width: 20px;
        border-radius: 6px;
        border: 1px solid {COLORS["border"]};
    }}

    QScrollBar::add-line:horizontal, QScrollBar::sub-line:horizontal {{
        width: 0px;
    }}

    QScrollBar::add-line:vertical, QScrollBar::sub-line:vertical {{
        height: 0px;
    }}

    QTableView QTableCornerButton::section {{
        background-color: {COLORS["panel"]};
        border: 1px solid {COLORS["border"]};
    }}

    QTableView QHeaderView {{
        background-color: {COLORS["panel"]};
    }}

    QDialog,
    QMessageBox,
    QInputDialog {{
        background-color: {COLORS["surface"]};
        color: {COLORS["text"]};
    }}

    QDialog QLabel,
    QMessageBox QLabel,
    QInputDialog QLabel {{
        color: {COLORS["text"]};
    }}

    QDialog QLineEdit,
    QMessageBox QLineEdit,
    QInputDialog QLineEdit {{
        background-color: {COLORS["surface_alt"]};
        color: {COLORS["text"]};
        border: 1px solid {COLORS["border"]};
        border-radius: 6px;
        padding: 6px 8px;
    }}

    QDialog QLineEdit:focus,
    QMessageBox QLineEdit:focus,
    QInputDialog QLineEdit:focus {{
        border-color: {COLORS["accent"]};
    }}

    QDialogButtonBox QPushButton {{
        background-color: {COLORS["panel"]};
        border: 1px solid {COLORS["border"]};
        border-radius: 8px;
        padding: 8px 12px;
        color: {COLORS["text"]};
        font: 600 13px;
    }}

    QDialogButtonBox QPushButton:hover {{
        background-color: {COLORS["accent_strong"]};
        border-color: {COLORS["accent"]};
    }}
"""

WIZARD_QSS = f"""
    QWizard {{
        background-color: {COLORS["bg"]};
        color: {COLORS["text"]};
    }}

    QWizardPage {{
        background-color: {COLORS["surface"]};
        color: {COLORS["text"]};
    }}

    QLabel {{
        color: {COLORS["text"]};
    }}

    QLineEdit,
    QComboBox {{
        background-color: {COLORS["surface_alt"]};
        color: {COLORS["text"]};
        border: 1px solid {COLORS["border"]};
        border-radius: 6px;
        padding: 6px 8px;
    }}

    QLineEdit:focus,
    QComboBox:focus {{
        border-color: {COLORS["accent"]};
    }}

    QComboBox::drop-down {{
        subcontrol-origin: padding;
        subcontrol-position: top right;
        width: 22px;
        border-left: 1px solid {COLORS["border"]};
        background-color: {COLORS["panel"]};
        border-top-right-radius: 6px;
        border-bottom-right-radius: 6px;
    }}

    QComboBox::down-arrow {{
        image: none;
        width: 0px;
        height: 0px;
        border-left: 5px solid transparent;
        border-right: 5px solid transparent;
        border-top: 6px solid {COLORS["text_muted"]};
    }}

    QComboBox QAbstractItemView {{
        background-color: {COLORS["surface_alt"]};
        color: {COLORS["text"]};
        selection-background-color: {COLORS["accent_soft"]};
        selection-color: {COLORS["text"]};
        border: 1px solid {COLORS["border"]};
        outline: 0;
    }}

    QToolButton {{
        background-color: {COLORS["panel"]};
        border: 1px solid {COLORS["border"]};
        border-radius: 6px;
        padding: 4px;
    }}

    QToolButton:hover {{
        background-color: {COLORS["accent_strong"]};
        border-color: {COLORS["accent"]};
    }}

    QGroupBox {{
        color: {COLORS["text"]};
        border: 1px solid {COLORS["border"]};
        border-radius: 8px;
        margin-top: 10px;
    }}

    QGroupBox::title {{
        subcontrol-origin: margin;
        subcontrol-position: top left;
        padding: 2px 6px;
        color: {COLORS["text_muted"]};
    }}

    QCheckBox {{
        color: {COLORS["text"]};
    }}

    QCheckBox::indicator {{
        width: 14px;
        height: 14px;
        border-radius: 4px;
        border: 1px solid {COLORS["border"]};
        background-color: {COLORS["surface_alt"]};
    }}

    QCheckBox::indicator:checked {{
        background-color: {COLORS["accent"]};
        border-color: {COLORS["accent"]};
    }}

    QMenu {{
        background-color: {COLORS["menu_bg"]};
        color: {COLORS["text"]};
        border: 1px solid {COLORS["border"]};
    }}

    QMenu::item::selected {{
        background-color: {COLORS["panel_alt"]};
    }}

    QToolTip {{
        background-color: {COLORS["tooltip"]};
        border: 1px solid {COLORS["border"]};
        color: {COLORS["tooltip_text"]};
        padding: 4px 6px;
        opacity: 220;
    }}

    QTableView {{
        background-color: {COLORS["surface_alt"]};
        alternate-background-color: {COLORS["surface"]};
        color: {COLORS["text"]};
        border: 1px solid {COLORS["border"]};
        border-radius: 6px;
        gridline-color: {COLORS["border"]};
    }}

    QHeaderView::section {{
        background-color: {COLORS["panel"]};
        color: {COLORS["text"]};
        padding: 6px;
        border: 1px solid {COLORS["border"]};
    }}

    QTableView::item {{
        padding: 8px;
        color: {COLORS["text"]};
    }}

    QTableView::item:selected {{
        selection-background-color: {COLORS["accent_soft"]};
        selection-color: {COLORS["text"]};
    }}

    QMessageBox {{
        background-color: {COLORS["surface"]};
        color: {COLORS["text"]};
    }}

    QMessageBox QLabel {{
        color: {COLORS["text"]};
    }}

    QMessageBox QPushButton {{
        background-color: {COLORS["panel"]};
        border: 1px solid {COLORS["border"]};
        border-radius: 8px;
        padding: 8px 12px;
        color: {COLORS["text"]};
        font: 600 13px;
    }}

    QMessageBox QPushButton:hover {{
        background-color: {COLORS["accent_strong"]};
        border-color: {COLORS["accent"]};
    }}
"""

PACKAGE_DIALOG_QSS = f"""
    QDialog {{
        background-color: {COLORS["surface"]};
        color: {COLORS["text"]};
    }}

    QLabel {{
        color: {COLORS["text"]};
    }}

    QLineEdit {{
        background-color: {COLORS["surface_alt"]};
        color: {COLORS["text"]};
        border: 1px solid {COLORS["border"]};
        border-radius: 6px;
        padding: 6px 8px;
    }}

    QLineEdit:focus {{
        border-color: {COLORS["accent"]};
    }}

    QFrame[frameShape="4"] {{
        color: {COLORS["border"]};
        border: 1px solid {COLORS["border"]};
    }}

    QTableView {{
        background-color: {COLORS["surface_alt"]};
        alternate-background-color: {COLORS["surface"]};
        color: {COLORS["text"]};
        border: 1px solid {COLORS["border"]};
        gridline-color: {COLORS["border"]};
        border-radius: 6px;
    }}

    QHeaderView::section {{
        background-color: {COLORS["panel"]};
        color: {COLORS["text"]};
        padding: 6px;
        border: 1px solid {COLORS["border"]};
    }}

    QTableView::item:selected {{
        background-color: {COLORS["accent_soft"]};
        color: {COLORS["text"]};
    }}

    QMenu {{
        background-color: {COLORS["menu_bg"]};
        color: {COLORS["text"]};
        border: 1px solid {COLORS["border"]};
    }}

    QMenu::item:selected {{
        background-color: {COLORS["panel_alt"]};
    }}

    QMessageBox {{
        background-color: {COLORS["surface"]};
        color: {COLORS["text"]};
    }}

    QMessageBox QLabel {{
        color: {COLORS["text"]};
    }}

    QMessageBox QPushButton {{
        background-color: {COLORS["panel"]};
        border: 1px solid {COLORS["border"]};
        border-radius: 8px;
        padding: 8px 12px;
        color: {COLORS["text"]};
        font: 600 13px;
    }}

    QMessageBox QPushButton:hover {{
        background-color: {COLORS["accent_strong"]};
        border-color: {COLORS["accent"]};
    }}
"""

DIALOG_QSS = f"""
    QDialog {{
        background-color: {COLORS["surface"]};
        color: {COLORS["text"]};
    }}

    QLabel {{
        color: {COLORS["text"]};
    }}

    QPushButton {{
        background-color: {COLORS["panel"]};
        border: 1px solid {COLORS["border"]};
        border-radius: 8px;
        padding: 8px 12px;
        color: {COLORS["text"]};
        font: 600 13px;
    }}

    QPushButton:hover {{
        background-color: {COLORS["accent_strong"]};
        border-color: {COLORS["accent"]};
    }}

    QProgressBar {{
        background-color: {COLORS["surface_alt"]};
        border: 1px solid {COLORS["border"]};
        border-radius: 6px;
        text-align: center;
        color: {COLORS["text"]};
    }}

    QProgressBar::chunk {{
        background-color: {COLORS["accent"]};
        border-radius: 6px;
    }}
"""



#    VenviPy - A Virtual Environment Manager for Python.
#    Copyright (C) 2021 - Youssef Serestou - sinusphi.sq@gmail.com
#
#    This program is free software: you can redistribute it and/or modify
#    it under the terms of the GNU General Public License as published by
#    the Free Software Foundation, either version 3 of the License or any
#    later version.
#
#    This program is distributed in the hope that it will be useful,
#    but WITHOUT ANY WARRANTY; without even the implied warranty of
#    MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
#    GNU General Public License for more details.
#
#    A copy of the GNU General Public License version 3 named LICENSE is
#    in the root directory of VenviPy.
#    If not, see <https://www.gnu.org/licenses/licenses.en.html#GPL>.

# -*- coding: utf-8 -*-
"""
The main module of VenviPy.
"""
import sys
import os
import csv
import getopt
import logging
from functools import partial
from pathlib import Path

# need to set the correct cwd
CURRENT_DIR = Path(__file__).parent
sys.path.insert(0, str(CURRENT_DIR))
os.chdir(CURRENT_DIR)

from PyQt6 import QtCore
from PyQt6.QtCore import Qt, QRect, QSize, pyqtSlot
from PyQt6.QtGui import (
    QIcon,
    QPixmap,
    QStandardItemModel,
    QStandardItem,
    QAction
)
from PyQt6.QtWidgets import (
    QStyle,
    QMainWindow,
    QApplication,
    QFileDialog,
    QLabel,
    QToolButton,
    QWidget,
    QGridLayout,
    QVBoxLayout,
    QPushButton,
    QSpacerItem,
    QSizePolicy,
    QInputDialog,
    QCheckBox,
    QMenuBar,
    QMenu,
    QStatusBar,
    QAbstractItemView,
    QMessageBox,
    QHBoxLayout,
    QLineEdit,
    QTabWidget,
    QWidgetAction
)

import venvipy_rc  # pylint: disable=unused-import
import get_data
import wizard
import bars
from styles import theme
from styles import custom
from pkg_installer import PackageInstaller
from pkg_manager import PackageManager
from dialogs import InfoAboutVenviPy
from tables import VenvTable, InterpreterTable


LOG_FORMAT = "[%(levelname)s] - { %(name)s }: %(message)s"
logger = logging.getLogger()



class MainWindow(QMainWindow):
    """
    The main window.
    """
    def __init__(self):
        super().__init__()

        self.init_ui()


    def init_ui(self):
        self.setWindowFlag(Qt.WindowType.FramelessWindowHint)
        self.setWindowTitle("VenviPy")
        self.resize(1600, 880)
        self.setContentsMargins(4, 0, 4, 0)
        self.center()
        self.setWindowIcon(QIcon(":/img/profile.png"))
        self.setStyleSheet(theme.dark)

        self.info_about_venvipy = InfoAboutVenviPy()
        self.venv_wizard = wizard.VenvWizard()
        self._wizard_refresh_tab = None

        # refresh venv table on wizard close
        self.venv_wizard.refresh.connect(self.refresh_wizard_tab)

        # refresh interpreter table if 'py-installs' changes
        self.venv_wizard.update_table.connect(self.pop_interpreter_table)

        self.pkg_installer = PackageInstaller()
        self.pkg_manager = PackageManager()


        #]===================================================================[#
        #] ICONS [#==========================================================[#
        #]===================================================================[#

        find_icon = QIcon.fromTheme("edit-find")
        #python_icon = QIcon(":/img/python.png")
        #manage_icon = QIcon.fromTheme("insert-object")
        #settings_icon = QIcon.fromTheme("preferences-system")

        new_icon = QIcon(
            self.style().standardIcon(QStyle.StandardPixmap.SP_FileDialogNewFolder)
        )
        exit_icon = QIcon(
            self.style().standardIcon(QStyle.StandardPixmap.SP_BrowserStop)
        )
        folder_icon = QIcon(
            self.style().standardIcon(QStyle.StandardPixmap.SP_DirOpenIcon)
        )
        qt_icon = QIcon(
            self.style().standardIcon(QStyle.StandardPixmap.SP_TitleBarMenuButton)
        )
        info_icon = QIcon(
            self.style().standardIcon(QStyle.StandardPixmap.SP_FileDialogInfoView)
        )
        self.new_folder_icon = QIcon(
            self.style().standardIcon(QStyle.StandardPixmap.SP_FileDialogNewFolder)
        )
        self.reload_icon = QIcon(
            self.style().standardIcon(QStyle.StandardPixmap.SP_BrowserReload)
        )
        self.dir_open_icon = QIcon(
            self.style().standardIcon(QStyle.StandardPixmap.SP_DirOpenIcon)
        )
        #delete_icon = QIcon(
        #    self.style().standardIcon(QStyle.StandardPixmap.SP_TrashIcon)
        #)


        #]===================================================================[#
        #] LAYOUTS [#========================================================[#
        #]===================================================================[#

        centralwidget = QWidget(self)
        grid_layout = QGridLayout(centralwidget)

        v_layout_1 = QVBoxLayout()
        v_layout_2 = QVBoxLayout()
        v_layout_1.setContentsMargins(12, 19, 5, -1)
        v_layout_2.setContentsMargins(-1, 4, 6, -1)

        # python logo
        self.logo = QLabel(centralwidget)
        self.logo.setPixmap(QPixmap(":/img/pypower.png"))
        self.logo.setAlignment(Qt.AlignmentFlag.AlignRight)


        #]===================================================================[#
        #] BUTTONS [#========================================================[#
        #]===================================================================[#

        self.add_interpreter_button = QPushButton(
            "Add &Interpreter",
            centralwidget,
            statusTip="Add an Interpreter",
            clicked=self.add_interpreter
        )
        self.add_interpreter_button.setMinimumSize(QSize(150, 0))

        self.new_venv_button = QPushButton(
            "&New Venv",
            centralwidget,
            statusTip="Create a new virtual environment",
            clicked=self.launch_venv_wizard
        )
        self.new_venv_button.setMinimumSize(QSize(135, 0))

        self.exit_button = QPushButton(
            "Quit",
            centralwidget,
            statusTip="Quit Application",
            clicked=self.on_close
        )

        # use line edit to store the str
        self.directory_line = QLineEdit()
        self.venv_tab_map = {}
        self.venv_tabs_data = []
        self.tab_save_pref = {
            "always_save_tabs": False,
            "ask_before_saving_tabs": True
        }

        #]===================================================================[#
        # spacer between manage button and exit button
        spacer_item_1 = QSpacerItem(
            20, 40, QSizePolicy.Policy.Minimum, QSizePolicy.Policy.Expanding
        )
        #]===================================================================[#

        v_layout_2.addWidget(self.logo)
        v_layout_2.addWidget(self.add_interpreter_button)
        v_layout_2.addWidget(self.new_venv_button)
        v_layout_2.addItem(spacer_item_1)
        v_layout_2.addWidget(self.exit_button)

        grid_layout.addLayout(v_layout_2, 0, 1, 1, 1)


        #]===================================================================[#
        #] TABLES [#=========================================================[#
        #]===================================================================[#

        # interpreter table header
        interpreter_table_title = QLabel(
            custom.interpreter_table_title_text,
            centralwidget
        )

        # interpreter table
        self.interpreter_table = InterpreterTable(
            centralwidget,
            selectionBehavior=QAbstractItemView.SelectionBehavior.SelectRows,
            editTriggers=QAbstractItemView.EditTrigger.NoEditTriggers,
            alternatingRowColors=True,
            sortingEnabled=True,
            drop_item=self.pop_interpreter_table
        )

        # hide vertical header
        self.interpreter_table.verticalHeader().hide()

        # adjust (horizontal) headers
        h_header_interpreter_table = self.interpreter_table.horizontalHeader()
        h_header_interpreter_table.setDefaultAlignment(Qt.AlignmentFlag.AlignLeft)
        h_header_interpreter_table.setDefaultSectionSize(180)
        h_header_interpreter_table.setStretchLastSection(True)

        # set table view model
        self.model_interpreter_table = QStandardItemModel(0, 2, centralwidget)
        self.model_interpreter_table.setHorizontalHeaderLabels(
            ["Version", "Path"]
        )
        self.interpreter_table.setModel(self.model_interpreter_table)

        #]===================================================================[#
        # spacer between interpreter table and venv table title
        spacer_item_2 = QSpacerItem(
            20, 12, QSizePolicy.Policy.Minimum, QSizePolicy.Policy.Fixed
        )
        #]===================================================================[#

        self.venv_tabs = QTabWidget(centralwidget)
        self.venv_tabs.currentChanged.connect(self.on_venv_tab_changed)
        self.venv_tabs.setTabsClosable(True)
        self.venv_tabs.tabCloseRequested.connect(self.close_venv_tab)
        self.venv_tabs.tabBarDoubleClicked.connect(self.rename_venv_tab)
        tab_bar = self.venv_tabs.tabBar()
        tab_bar.setMovable(True)
        tab_bar.tabMoved.connect(self.on_venv_tab_moved)

        # add widgets to layout
        v_layout_1.addWidget(interpreter_table_title)
        v_layout_1.addWidget(self.interpreter_table)
        v_layout_1.addItem(spacer_item_2)
        v_layout_1.addWidget(self.venv_tabs)

        grid_layout.addLayout(v_layout_1, 0, 0, 1, 1)

        self.setCentralWidget(centralwidget)
        restored = self.restore_tabs_from_state()
        if not restored:
            initial_dir = get_data.get_active_dir_str()
            initial_title = self.next_venv_tab_title(initial_dir)
            self.create_venv_tab(initial_title, initial_dir)

        self.on_venv_tab_changed(self.venv_tabs.currentIndex())


        #]===================================================================[#
        #] ACTIONS [#========================================================[#
        #]===================================================================[#

        # create actions
        self.action_add_interpreter = QAction(
            find_icon,
            "Add &Interpreter",
            self,
            statusTip="Add an Interpreter",
            shortcut="Ctrl+I",
            triggered=self.add_interpreter
        )

        self.action_new_venv = QAction(
            new_icon,
            "&New Venv",
            self,
            statusTip="Create a new virtual environment",
            shortcut="Ctrl+N",
            triggered=self.launch_venv_wizard
        )

        self.action_select_active_dir = QAction(
            folder_icon,
            "Change active &directory",
            self,
            statusTip="Change active directory",
            shortcut="Ctrl+D",
            triggered=self.select_active_dir
        )

        self.always_save_tabs_checkbox = QCheckBox("Always save tabs", self)
        self.always_save_tabs_checkbox.setChecked(
            self.tab_save_pref.get("always_save_tabs", False)
        )
        self.always_save_tabs_checkbox.toggled.connect(
            self.toggle_always_save_tabs
        )
        self.action_always_save_tabs = QWidgetAction(self)
        self.action_always_save_tabs.setDefaultWidget(
            self.always_save_tabs_checkbox
        )

        self.action_exit = QAction(
            exit_icon,
            "&Quit",
            self,
            statusTip="Quit application",
            shortcut="Ctrl+Q",
            triggered=self.on_close
        )

        self.action_about_venvipy = QAction(
            info_icon,
            "&About VenviPy",
            self,
            statusTip="About VenviPy",
            shortcut="Ctrl+A",
            triggered=self.info_about_venvipy.exec
        )

        self.action_about_qt = QAction(
            qt_icon,
            "About &Qt",
            self,
            statusTip="About Qt",
            shortcut="Ctrl+Q",
            triggered=self.info_about_qt
        )


        #]===================================================================[#
        #] MENUS [#==========================================================[#
        #]===================================================================[#

        status_bar = QStatusBar(self)
        self.setStatusBar(status_bar)

        menu_bar = QMenuBar(self)

        self.title_bar = bars.TitleBar(self)
        menu_container = QWidget(self)
        menu_container.setObjectName("menuContainer")

        menu_layout = QVBoxLayout(menu_container)
        menu_layout.setContentsMargins(2, 0, 2, 0)
        menu_layout.setSpacing(2)

        title_separator = QWidget(self)
        title_separator.setObjectName("titleSeparator")
        title_separator.setFixedHeight(1)

        menu_layout.addWidget(self.title_bar)
        menu_layout.addWidget(title_separator)
        menu_layout.addWidget(menu_bar)

        self.setMenuWidget(menu_container)

        menu_venv = QMenu("&Venv", menu_bar)
        menu_venv.addAction(self.action_add_interpreter)
        menu_venv.addSeparator()
        menu_venv.addAction(self.action_new_venv)
        menu_venv.addAction(self.action_select_active_dir)
        menu_venv.addAction(self.action_always_save_tabs)
        menu_venv.addSeparator()
        menu_venv.addAction(self.action_exit)
        menu_bar.addAction(menu_venv.menuAction())

        menu_help = QMenu("&Help", menu_bar)
        menu_help.addAction(self.action_about_qt)
        menu_help.addAction(self.action_about_venvipy)
        menu_bar.addAction(menu_help.menuAction())

        # check if any Python is installed
        if os.path.exists(get_data.DB_FILE):
            with open(get_data.DB_FILE, "r", encoding="utf-8") as f:
                lines = f.readlines()
            if len(lines) < 2:
                self.launching_without_python()


    def launching_without_python(self):
        """If no Python was found run with features disabled.
        """
        logger.warning("No suitable Python interpreter found")
        msg_txt = (
            "No suitable Python interpreter found!\n\n"
            "Please specify the path to a Python (>=3.6) \n"
            "interpreter or click 'Continue' to go on anyway.\n\n"
        )
        self.msg_box = QMessageBox(
            QMessageBox.Icon.Critical,
            "VenviPy",
            msg_txt, QMessageBox.StandardButton.NoButton,
            self
        )
        select_button = self.msg_box.addButton(
            "&Select", QMessageBox.ButtonRole.AcceptRole
        )
        self.msg_box.addButton("&Continue", QMessageBox.ButtonRole.RejectRole)

        self.msg_box.exec()
        if self.msg_box.clickedButton() == select_button:
            # let user specify path to an interpreter
            self.add_interpreter()
        else:
            self.enable_features(False)


    def center(self):
        """Center window.
        """
        qr = self.frameGeometry()
        screen = self.screen() or QApplication.primaryScreen()
        if screen:
            cp = screen.availableGeometry().center()
            qr.moveCenter(cp)
            self.move(qr.topLeft())


    def on_close(self):
        """Request application shutdown.
        """
        self.close()


    def closeEvent(self, event):
        """Handle persistence and shutdown when the window closes.
        """
        if not self.handle_tab_persistence_on_close():
            event.ignore()
            return

        self.shutdown_threads()
        event.accept()
        super().closeEvent(event)


    def shutdown_threads(self):
        """Stop background threads gracefully.
        """
        if hasattr(self.venv_wizard, "basic_settings"):
            thread = getattr(self.venv_wizard.basic_settings, "thread", None)
            if thread is not None:
                thread.exit()

        for tab_data in self.venv_tabs_data:
            table = tab_data.get("table")
            if table is not None and hasattr(table, "thread"):
                table.thread.exit()


    def handle_tab_persistence_on_close(self):
        """Ask whether to save tabs (or auto-save) when closing.
        """
        tab_count = self.venv_tabs.count()

        if self.tab_save_pref.get("always_save_tabs"):
            self.save_tabs_state()
            return True

        if not self.tab_save_pref.get("ask_before_saving_tabs", True):
            # Safety net: default to saving when prompting is disabled.
            self.save_tabs_state()
            return True

        # Only prompt if there is at least one tab open.
        if tab_count == 0:
            return True
        # Skip prompt when no tab has a directory loaded.
        any_dir_loaded = False
        for i in range(tab_count):
            tab_widget = self.venv_tabs.widget(i)
            tab_data = self.get_tab_data(tab_widget)
            if tab_data and tab_data.get("path"):
                any_dir_loaded = True
                break
        if not any_dir_loaded:
            return True

        msg_box = QMessageBox(
            QMessageBox.Icon.Question,
            "Save tabs?",
            (
                f"There {'is' if tab_count == 1 else 'are'} "
                f"{tab_count} open tab{'s' if tab_count != 1 else ''}.\n"
                "Save them for the next start?"
            ),
            QMessageBox.StandardButton.Yes
            | QMessageBox.StandardButton.No
            | QMessageBox.StandardButton.Cancel,
            self
        )

        dont_ask_checkbox = QCheckBox("Don't ask again. Always save tabs.")
        msg_box.setCheckBox(dont_ask_checkbox)

        result = msg_box.exec()

        if result == QMessageBox.StandardButton.Cancel:
            return False

        if result == QMessageBox.StandardButton.No:
            state = self.collect_tabs_state()
            state["tabs"] = []
            state["active_index"] = 0
            get_data.save_tabs_state(state)
            self.set_active_dir("")
            return True

        if dont_ask_checkbox.isChecked():
            self.tab_save_pref["always_save_tabs"] = True
            self.tab_save_pref["ask_before_saving_tabs"] = False

        if result == QMessageBox.StandardButton.Yes:
            self.save_tabs_state()

        return True


    def toggle_always_save_tabs(self, enabled):
        """Toggle automatic tab persistence on app close.
        """
        self.tab_save_pref["always_save_tabs"] = bool(enabled)
        self.tab_save_pref["ask_before_saving_tabs"] = not bool(enabled)
        self.save_tabs_state()


    def changeEvent(self, event):
        super().changeEvent(event)
        if event.type() == QtCore.QEvent.Type.WindowStateChange:
            if hasattr(self, "title_bar"):
                self.title_bar.update_maximize_icon()


    def info_about_qt(self):
        """Open the "About Qt" dialog.
        """
        QMessageBox.aboutQt(self)


    def add_interpreter(self):
        """Add a custom interpreter.
        """
        if self.venv_wizard.basic_settings.select_python() != "":
            self.enable_features(True)


    def enable_features(self, state):
        """Enable or disable features.
        """
        for tab_data in self.venv_tabs_data:
            tab_data["table"].setEnabled(state)


    @pyqtSlot()
    def pop_interpreter_table(self):
        """Populate the interpreter table view.
        """
        get_data.ensure_dbfile()

        with open(get_data.DB_FILE, newline="", encoding="utf-8") as cf:
            reader = csv.DictReader(cf, delimiter=",")
            self.model_interpreter_table.setRowCount(0)
            for info in reader:
                self.model_interpreter_table.insertRow(0)
                for i, text in enumerate(
                        (info["PYTHON_VERSION"], info["PYTHON_PATH"])
                    ):
                    self.model_interpreter_table.setItem(
                        0, i, QStandardItem(text)
                    )
        # also populate the combo box in wizard
        self.venv_wizard.basic_settings.pop_combo_box()


    def pop_venv_table(self, tab_widget=None):
        """Populate the venv table view.
        """
        tab_data = self.get_tab_data(tab_widget)
        if tab_data is None:
            return

        model = tab_data["model"]
        model.setRowCount(0)

        for info in get_data.get_venvs(tab_data["path"]):
            model.insertRow(0)
            for i, text in enumerate((
                    info.venv_name,
                    info.venv_version,
                    info.site_packages,
                    info.is_installed,
                    info.venv_comment
            )):
                model.setItem(0, i, QStandardItem(text))

        model.sort(0, Qt.SortOrder.AscendingOrder)


    def update_label(self, tab_data=None):
        """
        Show the currently selected folder containing
        virtual environments.
        """
        tab_data = self.get_tab_data(tab_data)
        if tab_data is None:
            return

        head = (
            '<span style="font-size: 13pt;">\
                    <b>Virtual environments:</b>\
                </span>'
        )
        no_folder = (
            '<span style="font-size: 13pt; color: #ff0000;">\
                <i> --- please select a folder containing virtual environments</i>\
            </span>'
        )
        with_folder = (
            f'<span style="font-size: 13pt; color: #0059ff;">\
                {tab_data["path"]}\
            </span>'
        )

        if tab_data["path"] != "":
            tab_data["label"].setText(f"{head}{with_folder}")
        else:
            tab_data["label"].setText(f"{head}{no_folder}")


    def select_active_dir(self, tab_widget=None):
        """
        Select the active directory of which the content
        should be shown in venv table.
        """
        tab_data = self.get_tab_data(tab_widget)
        if tab_data is None:
            return

        current_dir = tab_data["path"] or get_data.get_active_dir_str()

        directory = QFileDialog.getExistingDirectory(
            self,
            "Open a folder containing virtual environments",
            directory=current_dir
        )
        self.directory_line.setText(directory)
        active_dir = self.directory_line.text()

        if active_dir != "":
            tab_data["path"] = active_dir
            self.set_active_dir(active_dir)
            self.pop_venv_table(tab_data["widget"])
            self.update_label(tab_data)


    def add_venv_tab(self):
        """Create a new venv tab for another directory.
        """
        current_dir = get_data.get_active_dir_str()
        directory = QFileDialog.getExistingDirectory(
            self,
            "Open a folder containing virtual environments",
            directory=current_dir
        )
        if directory != "":
            if self.venv_tabs.count() == 1:
                current_data = self.get_tab_data()
                if current_data and not current_data.get("path"):
                    title = self.next_venv_tab_title(directory)
                    current_data["path"] = directory
                    current_data["title"] = title
                    current_index = self.venv_tabs.currentIndex()
                    if current_index >= 0:
                        self.venv_tabs.setTabText(current_index, title)
                    self.set_active_dir(directory)
                    self.pop_venv_table(current_data["widget"])
                    self.update_label(current_data)
                    return

            title = self.next_venv_tab_title(directory)
            tab_data = self.create_venv_tab(title, directory)
            self.venv_tabs.setCurrentWidget(tab_data["widget"])
            self.set_active_dir(directory)


    def create_venv_tab(self, title, active_dir):
        """Create a new venv tab.
        """
        tab_widget = QWidget(self)
        tab_layout = QVBoxLayout(tab_widget)
        header_layout = QHBoxLayout()

        label = QLabel(tab_widget)
        new_venv_button = QToolButton(
            text="+",
            toolTip="New venv",
            statusTip="Create a new virtual environment",
            clicked=self.launch_venv_wizard
        )
        new_venv_button.setFixedSize(30, 30)

        add_tab_button = QToolButton(
            icon=self.new_folder_icon,
            toolTip="Add a venv tab",
            statusTip="Open a new venv tab",
            clicked=self.add_venv_tab
        )
        add_tab_button.setFixedSize(30, 30)

        reload_button = QToolButton(
            icon=self.reload_icon,
            toolTip="Reload",
            statusTip="Reload venv table content",
            clicked=partial(self.pop_venv_table, tab_widget)
        )
        reload_button.setFixedSize(30, 30)

        active_dir_button = QToolButton(
            icon=self.dir_open_icon,
            toolTip="Switch directory",
            statusTip="Select another directory",
            clicked=partial(self.select_active_dir, tab_widget)
        )
        active_dir_button.setFixedSize(30, 30)

        header_layout.addWidget(label)
        header_layout.addStretch(1)
        header_layout.addWidget(new_venv_button)
        header_layout.addWidget(add_tab_button)
        header_layout.addWidget(reload_button)
        header_layout.addWidget(active_dir_button)

        venv_table = VenvTable(
            tab_widget,
            selectionBehavior=QAbstractItemView.SelectionBehavior.SelectRows,
            editTriggers=QAbstractItemView.EditTrigger.NoEditTriggers,
            alternatingRowColors=True,
            sortingEnabled=True,
            refresh=partial(self.pop_venv_table, tab_widget),
            start_installer=self.pkg_installer.launch,
            start_pkg_manager=self.pkg_manager.launch
        )

        # hide vertical header
        venv_table.verticalHeader().hide()

        # adjust horizontal headers
        h_header_venv_table = venv_table.horizontalHeader()
        h_header_venv_table.setDefaultAlignment(Qt.AlignmentFlag.AlignLeft)
        h_header_venv_table.setStretchLastSection(True)

        # set table view model
        model_venv_table = QStandardItemModel(0, 4, tab_widget)
        model_venv_table.setHorizontalHeaderLabels([
            "Venv",
            "Version",
            "Packages",
            "Installed",
            "Description"
        ])
        venv_table.setModel(model_venv_table)

        # adjust column width
        venv_table.setColumnWidth(0, 225)
        venv_table.setColumnWidth(1, 120)
        venv_table.setColumnWidth(2, 100)
        venv_table.setColumnWidth(3, 80)

        tab_layout.addLayout(header_layout)
        tab_layout.addWidget(venv_table)

        self.venv_tabs.addTab(tab_widget, title)
        tab_data = {
            "widget": tab_widget,
            "label": label,
            "table": venv_table,
            "model": model_venv_table,
            "path": active_dir,
            "title": title
        }
        self.venv_tabs_data.append(tab_data)
        self.venv_tab_map[tab_widget] = tab_data

        self.update_label(tab_data)
        self.pop_venv_table(tab_widget)

        return tab_data


    def restore_tabs_from_state(self):
        """Recreate tabs based on the last saved state, if any.
        """
        state = get_data.load_tabs_state() or {}

        self.tab_save_pref["always_save_tabs"] = bool(
            state.get("always_save_tabs", False)
        )
        self.tab_save_pref["ask_before_saving_tabs"] = bool(
            state.get("ask_before_saving_tabs", True)
        )
        if not self.tab_save_pref["ask_before_saving_tabs"]:
            self.tab_save_pref["always_save_tabs"] = True

        tabs = state.get("tabs") or []
        if not tabs:
            return False

        for tab in tabs:
            path = tab.get("path", "")
            title = tab.get("title") or self.next_venv_tab_title(path)
            self.create_venv_tab(title, path)

        active_index = state.get("active_index", 0)
        if 0 <= active_index < self.venv_tabs.count():
            self.venv_tabs.setCurrentIndex(active_index)

        return True


    def collect_tabs_state(self):
        """Gather current tab information for persistence.
        """
        tabs = []
        for i in range(self.venv_tabs.count()):
            tab_widget = self.venv_tabs.widget(i)
            tab_data = self.get_tab_data(tab_widget)
            if tab_data is None:
                continue
            tabs.append({
                "title": self.venv_tabs.tabText(i),
                "path": tab_data.get("path", "")
            })

        active_index = self.venv_tabs.currentIndex()
        if active_index < 0:
            active_index = 0

        return {
            "tabs": tabs,
            "active_index": active_index,
            "always_save_tabs": self.tab_save_pref.get("always_save_tabs", False),
            "ask_before_saving_tabs": self.tab_save_pref.get("ask_before_saving_tabs", True)
        }


    def save_tabs_state(self):
        """Persist current tabs and preferences.
        """
        state = self.collect_tabs_state()
        get_data.save_tabs_state(state)


    def rename_venv_tab(self, index):
        """Rename a venv tab via double-click.
        """
        if index < 0:
            return

        current_title = self.venv_tabs.tabText(index)
        new_title, ok = QInputDialog.getText(
            self,
            "Rename tab",
            "Tab name:",
            text=current_title
        )

        if not ok:
            return

        new_title = new_title.strip()
        if new_title == "":
            return

        self.venv_tabs.setTabText(index, new_title)

        tab_widget = self.venv_tabs.widget(index)
        tab_data = self.get_tab_data(tab_widget)
        if tab_data is not None:
            tab_data["title"] = new_title


    def on_venv_tab_moved(self, _from, _to):
        """Keep internal tab order in sync with the UI order.
        """
        new_order = []
        for i in range(self.venv_tabs.count()):
            tab_widget = self.venv_tabs.widget(i)
            tab_data = self.get_tab_data(tab_widget)
            if tab_data is not None:
                new_order.append(tab_data)

        self.venv_tabs_data = new_order


    def next_venv_tab_title(self, directory=None):
        """Return the default title for a new venv tab.
        """
        if directory:
            folder_name = Path(directory).name
            if folder_name:
                return folder_name
        return f"venv #{len(self.venv_tabs_data) + 1:02d}"


    def get_tab_data(self, tab_widget=None):
        """Return tab data for a widget (or the current tab).
        """
        if isinstance(tab_widget, dict):
            return tab_widget

        if tab_widget is None:
            tab_widget = self.venv_tabs.currentWidget()

        return self.venv_tab_map.get(tab_widget)


    def set_active_dir(self, active_dir):
        """Persist the active directory to disk.
        """
        get_data.ensure_active_dir()
        with open(get_data.ACTIVE_DIR, "w", encoding="utf-8") as f:
            f.write(active_dir)


    def on_venv_tab_changed(self, index):
        """Update active directory when switching tabs.
        """
        tab_widget = self.venv_tabs.widget(index)
        tab_data = self.get_tab_data(tab_widget)

        if tab_data is None:
            return

        self.set_active_dir(tab_data["path"])
        self.update_label(tab_data)


    def launch_venv_wizard(self):
        """Launch the venv wizard and remember the active tab.
        """
        tab_data = self.get_tab_data()
        self._wizard_refresh_tab = tab_data["widget"] if tab_data else None
        self.venv_wizard.exec()


    def refresh_wizard_tab(self):
        """Refresh the tab captured when launching the wizard.
        """
        tab_widget = self._wizard_refresh_tab

        if tab_widget is not None and tab_widget in self.venv_tab_map:
            self.pop_venv_table(tab_widget)
        else:
            self.pop_venv_table()

        self._wizard_refresh_tab = None


    def close_venv_tab(self, index):
        """Close a venv tab without confirmation.
        """
        if self.venv_tabs.count() <= 1:
            return

        tab_widget = self.venv_tabs.widget(index)

        if tab_widget is None:
            return

        tab_data = self.venv_tab_map.pop(tab_widget, None)

        if tab_data and tab_data in self.venv_tabs_data:
            self.venv_tabs_data.remove(tab_data)

        if tab_data:
            table = tab_data.get("table")

            if table and hasattr(table, "thread"):
                table.thread.quit()
                table.thread.wait()

            tab_widget.deleteLater()

        self.venv_tabs.removeTab(index)

        if self._wizard_refresh_tab == tab_widget:
            self._wizard_refresh_tab = None

        current_data = self.get_tab_data()

        if current_data:
            self.set_active_dir(current_data["path"])
            self.update_label(current_data)



def with_args():
    """Execute with command-line arguments.
    """
    # get full command-line arguments
    full_cmd_arguments = sys.argv

    # ignore the first
    argument_list = full_cmd_arguments[1:]

    # tell getopts() the parameters
    short_options = "Vdhwb"
    long_options = ["version", "debug", "help", "wizard", "wizard-debug"]

    try:
        arguments, values = getopt.getopt(
            argument_list, short_options, long_options
        )
    except getopt.error as e:
        # print error message and return error code
        err_msg = str(e)
        print(f"O{err_msg[1:]}")
        sys.exit(2)

    for arg, val in arguments:
        if arg in ("-h", "--help"):
            # show help message, then exit
            print(
                f"VenviPy v{get_data.__version__}  "
                "( https://github.com/sinusphi/venvipy )\n\n"
                "    -h --help           Print this help message and exit\n"
                "    -d --debug          Print debugging output\n"
                "    -v --version        Print version and exit\n"
                "    -w --wizard         Launch venv wizard only\n"
                "    -b --wizard-debug   Launch venv wizard with debug output\n"
            )
            sys.exit(0)

        if arg in ("-d", "--debug"):
            # verbose output for debugging
            logging.basicConfig(level=logging.DEBUG, format=LOG_FORMAT)

        if arg in ("-V", "--version"):
            # print version, then exit
            print(f"VenviPy {get_data.__version__}")
            sys.exit(0)

        if arg in ("-w", "--wizard"):
            os.system("clear")
            app = QApplication(sys.argv)
            venv_wizard = wizard.VenvWizard()
            venv_wizard.basic_settings.pop_combo_box()
            venv_wizard.exec()
            sys.exit(0)

        if arg in ("-b", "--wizard-debug"):
            os.system("clear")
            logging.basicConfig(level=logging.DEBUG, format=LOG_FORMAT)
            app = QApplication(sys.argv)
            venv_wizard = wizard.VenvWizard()
            venv_wizard.basic_settings.pop_combo_box()
            venv_wizard.exec()
            sys.exit(0)


def main():
    with_args()

    app = QApplication(sys.argv)
    os.system("clear")

    main_window = MainWindow()
    get_data.update_pypi_index()
    get_data.get_python_installs(True)
    main_window.pop_interpreter_table()
    main_window.venv_wizard.basic_settings.pop_combo_box()
    main_window.pop_venv_table()
    main_window.update_label()
    main_window.show()

    sys.exit(app.exec())



if __name__ == "__main__":
    main()
